# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#
import sys,os,glob,re
import ctypes
import ctypes.util
import types
import subprocess
import re
import tempfile
import urllib.request as urllib2
import tarfile
import shutil

try:
    from setuptools import setup, Extension, find_packages
    from setuptools.command.build_ext import build_ext
    from setuptools.command.build_clib import build_clib
    from setuptools.command.install import install
    from setuptools.command.install_scripts import install_scripts
    from setuptools.command.easy_install import easy_install
    from setuptools.command.test import test
    import setuptools
    print("Using setuptools version",setuptools.__version__)
except ImportError:
    print()
    print("****")
    print("    Installation requires setuptools version >= 38.")
    print("    Please upgrade or install with pip install -U setuptools")
    print("****")
    print()
    raise

# Turn this on for more verbose debugging output about compile attempts.
debug = False

print('Python version = ',sys.version)
py_version = "%d.%d"%sys.version_info[0:2]  # we check things based on the major.minor version.

scripts = ['galsim', 'galsim_download_cosmos']
scripts = [ os.path.join('bin',f) for f in scripts ]

def all_files_from(dir, ext=''):
    files = []
    for root, dirnames, filenames in os.walk(dir):
        for filename in filenames:
            if filename.endswith(ext) and not filename.startswith('.'):
                files.append(os.path.join(root, filename))
    return files

py_sources = all_files_from('pysrc', '.cpp')
cpp_sources = all_files_from('src', '.cpp')
test_sources = all_files_from('tests', '.cpp')
headers = all_files_from('include')
inst = all_files_from('src', '.inst')
shared_data = all_files_from('share')

copt =  {
    'gcc' : ['-O2','-std=c++11','-fvisibility=hidden','-fopenmp'],
    'gcc w/ GPU' : ['-O2','-std=c++11','-fvisibility=hidden','-fopenmp','-foffload=nvptx-none','-DGALSIM_USE_GPU'],
    'icc' : ['-O2','-vec-report0','-std=c++11','-openmp'],
    'clang' : ['-O2','-std=c++11',
               '-Wno-shorten-64-to-32','-fvisibility=hidden','-stdlib=libc++'],
    'clang w/ OpenMP' : ['-O2','-std=c++11','-fopenmp',
                         '-Wno-shorten-64-to-32','-fvisibility=hidden','-stdlib=libc++'],
    'clang w/ Intel OpenMP' : ['-O2','-std=c++11','-Xpreprocessor','-fopenmp',
                                '-Wno-shorten-64-to-32','-fvisibility=hidden','-stdlib=libc++'],
    'clang w/ manual OpenMP' : ['-O2','-std=c++11','-Xpreprocessor','-fopenmp',
                                '-Wno-shorten-64-to-32','-fvisibility=hidden','-stdlib=libc++'],
    'clang w/ GPU' : ['-O2','-msse2','-std=c++11','-fopenmp','-fopenmp-targets=nvptx64-nvidia-cuda',
                      '-Wno-openmp-mapping','-Wno-unknown-cuda-version',
                      '-Wno-shorten-64-to-32','-fvisibility=hidden', '-DGALSIM_USE_GPU'],
    'nvc++' : ['-O2','-std=c++11','-mp=gpu','-DGALSIM_USE_GPU'],
    'unknown' : [],
}
lopt =  {
    'gcc' : ['-fopenmp'],
    'gcc w/ GPU' : ['-fopenmp','-foffload=nvptx-none', '-foffload=-lm'],
    'icc' : ['-openmp'],
    'clang' : ['-stdlib=libc++'],
    'clang w/ OpenMP' : ['-stdlib=libc++','-fopenmp'],
    'clang w/ Intel OpenMP' : ['-stdlib=libc++','-liomp5'],
    'clang w/ manual OpenMP' : ['-stdlib=libc++','-lomp'],
    'clang w/ GPU' : ['-fopenmp','-fopenmp-targets=nvptx64-nvidia-cuda',
                      '-Wno-openmp-mapping','-Wno-unknown-cuda-version'],
    'nvc++' : ['-mp=gpu'],
    'unknown' : [],
}

# If we build with debug, undefine NDEBUG flag
# Note: setuptools stopped allowing --debug, so edit this manually if you want debugging.
# (The other debug variable is just for verbose output to debug this setup script.)
full_debug = False

undef_macros = []
if full_debug:
    undef_macros+=['NDEBUG']
    for name in copt.keys():
        if name != 'unknown':
            copt[name].append('-g')
        copt[name].append('-DMEM_TEST')
else:
    # Including mmgr.cpp in the library leads to problems if the other files don't
    # include mmgr.h.  So remove it.
    cpp_sources.remove('src/mmgr.cpp')

# Verbose is the default for setuptools logging, but if it's on the command line, we take it
# to mean that we should also be verbose.
if full_debug or "--verbose" in sys.argv:
    debug = True

local_tmp = 'tmp'

def get_compiler_type(compiler, check_unknown=True, output=False):
    """Try to figure out which kind of compiler this really is.
    In particular, try to distinguish between clang and gcc, either of which may
    be called cc or gcc.
    """
    if debug: output=True
    cc = compiler.compiler_so[0]
    if cc == 'ccache':
        cc = compiler.compiler_so[1]
    cmd = [cc,'--version']
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    lines = p.stdout.readlines()
    if output:
        print('compiler version information: ')
        for line in lines:
            print(line.decode().strip())
    # Python3 needs this decode bit.
    # Python2.7 doesn't need it, but it works fine.
    line = lines[0].decode(encoding='UTF-8')
    if line.startswith('Configured'):
        line = lines[1].decode(encoding='UTF-8')
    # nvc++ version info starts with a blank line
    if line.strip() == "":
        line = lines[1].decode(encoding='UTF-8')

    if 'clang' in line:
        # clang 3.7 is the first with openmp support.  But Apple lies about the version
        # number of clang, so the most reliable thing to do is to just try the compilation
        # with the openmp flag and see if it works.
        if output:
            print('Compiler is Clang.  Checking if it is a version that supports OpenMP.')
        if supports_gpu(compiler, 'clang w/ GPU'):
            if output:
                print("Yay! This version of clang supports GPU!")
            return 'clang w/ GPU'
        elif try_openmp(compiler, 'clang w/ OpenMP'):
            if output:
                print("Yay! This version of clang supports OpenMP!")
            return 'clang w/ OpenMP'
        elif try_openmp(compiler, 'clang w/ Intel OpenMP'):
            if output:
                print("Yay! This version of clang supports OpenMP!")
            return 'clang w/ Intel OpenMP'
        elif try_openmp(compiler, 'clang w/ manual OpenMP'):
            if output:
                print("Yay! This version of clang supports OpenMP!")
            return 'clang w/ manual OpenMP'
        else:
            if output:
                print("\nSorry.  This version of clang doesn't seem to support OpenMP.\n")
                print("If you think it should, you can use `python setup.py build --debug`")
                print("to get more information about the commands that failed.")
                print("You might need to add something to your C_INCLUDE_PATH or LIBRARY_PATH")
                print("(and probabaly LD_LIBRARY_PATH) to get it to work.\n")
            return 'clang'
    elif 'gcc' in line or 'GCC' in line:
        if supports_gpu(compiler, 'gcc w/ GPU'):
            if output:
                print("Yay! This version of gcc supports GPU!")
            return 'gcc w/ GPU'
        return 'gcc'
    elif 'nvc++' in line or 'nvcc' in line or 'NVIDIA' in line:
        return 'nvc++'
    elif 'clang' in cc:
        return 'clang'
    elif 'gcc' in cc or 'g++' in cc:
        return 'gcc'
    elif 'icc' in cc or 'icpc' in cc:
        return 'icc'
    elif check_unknown:
        # OK, the main thing we need to know is what openmp flag we need for this compiler,
        # so let's just try the various options and see what works.  Don't try icc, since
        # the -openmp flag there gets treated as '-o penmp' by gcc and clang, which is bad.
        # Plus, icc should be detected correctly by the above procedure anyway.
        if output:
            print('Unknown compiler.')
        for cc_type in ['gcc', 'clang w/ OpenMP', 'clang w/ manual OpenMP', 'clang w/ Intel OpenMP',
                        'clang']:
            if output:
                print('Check if the compiler works like ',cc_type)
            if try_openmp(compiler, cc_type):
                return cc_type
        # I guess none of them worked.  Now we really do have to bail.
        if output:
            print("None of these compile options worked.  Not adding any optimization flags.")
        return 'unknown'
    else:
        return 'unknown'

# Check whether this build of Clang supports offloading to GPU via OpenMP
def supports_gpu(compiler, cc_type):
    cpp_code = """
#include <iostream>

#define N 500
int main() {
    double arrA[N], arrB[N], arrC[N];

    for (int i = 0; i < N; i++) {
        arrA[i] = (double)i;
        arrB[i] = 1.0;
    }

#ifdef _OPENMP
#pragma omp target teams distribute parallel for
#endif
    for (int i = 0; i < N; i++) {
        arrC[i] = arrA[i] * arrB[i];
    }

    double sum = 0.0;
    for (int i = 0; i < N; i++) {
        sum += arrC[i];
    }
    std::cout << "Sum=" << sum << std::endl;
    return 0;
}
"""
    extra_cflags = copt[cc_type]
    extra_lflags = lopt[cc_type]
    return try_compile(cpp_code, compiler, extra_cflags, extra_lflags)

# Check for the fftw3 library in some likely places
def find_fftw_lib(output=False):
    import distutils.sysconfig

    if debug: output = True
    try_libdirs = []

    # Start with the explicit FFTW_DIR, if present.
    if 'FFTW_DIR' in os.environ:
        try_libdirs.append(os.environ['FFTW_DIR'])
        try_libdirs.append(os.path.join(os.environ['FFTW_DIR'],'lib'))

    # Add the python system library directory.
    try_libdirs.append(distutils.sysconfig.get_config_var('LIBDIR'))

    # If using Anaconda, add their lib dir in case fftw is installed there.
    # (With envs, this might be different than the sysconfig LIBDIR.)
    if 'CONDA_PREFIX' in os.environ:
        try_libdirs.append(os.path.join(os.environ['CONDA_PREFIX'],'lib'))

    # Try some standard locations where things get installed
    try_libdirs.extend(['/usr/local/lib', '/usr/lib'])
    if sys.platform == "darwin":
        try_libdirs.extend(['/sw/lib', '/opt/local/lib'])

    # Check the directories in LD_LIBRARY_PATH.  This doesn't work on OSX >= 10.11
    for path in ['LIBRARY_PATH', 'LD_LIBRARY_PATH', 'DYLD_LIBRARY_PATH']:
        if path in os.environ:
            for dir in os.environ[path].split(':'):
                try_libdirs.append(dir)

    # The user's home directory is often a good place to check.
    try_libdirs.append(os.path.join(os.path.expanduser("~"),"lib"))

    # If the above don't work, the fftw3 module may have the right directory.
    try:
        import fftw3
        try_libdirs.append(fftw3.lib.libdir)
    except ImportError:
        pass

    if sys.platform == "darwin":
        lib_ext = '.dylib'
    else:
        lib_ext = '.so'
    name = 'libfftw3' + lib_ext
    if output: print("Looking for ",name)
    tried_dirs = set()  # Keep track, so we don't try the same thing twice.
    for dir in try_libdirs:
        if dir == '': continue  # This messes things up if it's in there.
        if dir in tried_dirs: continue
        else: tried_dirs.add(dir)
        if not os.path.isdir(dir): continue
        libpath = os.path.join(dir, name)
        if not os.path.isfile(libpath): continue
        if output: print("  ", dir, end='')
        try:
            lib = ctypes.cdll.LoadLibrary(libpath)
            if output: print("  (yes)")
            return libpath
        except OSError as e:
            if output: print("  (no)")
            # Some places use lib64 rather than/in addition to lib.  Try that as well.
            if dir.endswith('lib') and os.path.isdir(dir + '64'):
                dir += '64'
                try:
                    libpath = os.path.join(dir, name)
                    if not os.path.isfile(libpath): continue
                    lib = ctypes.cdll.LoadLibrary(libpath)
                    if output: print("  ", dir, "  (yes)")
                    return libpath
                except OSError:
                    pass

    # If we didn't find it anywhere, but the user has set FFTW_DIR, trust it.
    if 'FFTW_DIR' in os.environ:
        libpath = os.path.join(os.environ['FFTW_DIR'], name)
        print("WARNING:")
        print("Could not find an installed fftw3 library named %s"%(name))
        print("Trusting the provided FFTW_DIR=%s for the library location."%(libpath))
        print("If this is incorrect, you may have errors later when linking.")
        return libpath

    # Last ditch attempt.  Use ctypes.util.find_library, which sometimes manages to find it
    # when the above attempts fail.
    try:
        libpath = ctypes.util.find_library('fftw3')
        if libpath == None:
            raise OSError
        if os.path.split(libpath)[0] == '':
            # If the above doesn't return a real path, try this instead.
            libpath = ctypes.util._findLib_gcc('fftw3')
            if libpath == None:
                raise OSError
        libpath = os.path.realpath(libpath)
        lib = ctypes.cdll.LoadLibrary(libpath)
    except Exception as e:
        print("Could not find fftw3 library.  Make sure it is installed either in a standard ")
        print("location such as /usr/local/lib, or the installation directory is either in ")
        print("your LIBRARY_PATH or FFTW_DIR environment variable.")
        raise
    else:
        dir, name = os.path.split(libpath)
        if output:
            if dir == '': dir = '[none]'
            print("  ", dir, "  (yes)")
        return libpath


# Check for Eigen in some likely places
def find_eigen_dir(output=False):
    if debug: output = True
    import distutils.sysconfig

    try_dirs = []

    # Start with a user-specified directory.
    if 'EIGEN_DIR' in os.environ:
        try_dirs.append(os.environ['EIGEN_DIR'])
        try_dirs.append(os.path.join(os.environ['EIGEN_DIR'], 'include'))

    # Add the python system include directory.
    try_dirs.append(distutils.sysconfig.get_config_var('INCLUDEDIR'))

    # If using Anaconda, add their lib dir in case fftw is installed there.
    # (With envs, this might be different than the sysconfig LIBDIR.)
    if 'CONDA_PREFIX' in os.environ:
        try_dirs.append(os.path.join(os.environ['CONDA_PREFIX'],'lib'))

    # Some standard install locations:
    try_dirs.extend(['/usr/local/include', '/usr/include'])
    if sys.platform == "darwin":
        try_dirs.extend(['/sw/include', '/opt/local/include'])

    # Also if there is a C_INCLUDE_PATH, check those dirs.
    for path in ['C_INCLUDE_PATH']:
        if path in os.environ:
            for dir in os.environ[path].split(':'):
                try_dirs.append(dir)

    # Finally, (last resort) check our own download of eigen.
    if os.path.isdir('downloaded_eigen'):
        try_dirs.extend(glob.glob(os.path.join('downloaded_eigen','*')))

    if output: print("Looking for Eigen:")
    for dir in try_dirs:
        if dir is None: continue
        if not os.path.isdir(dir): continue
        if output: print("  ", dir, end='')
        if os.path.isfile(os.path.join(dir, 'Eigen/Core')):
            if output: print("  (yes)")
            return dir
        if os.path.isfile(os.path.join(dir, 'eigen3', 'Eigen/Core')):
            dir = os.path.join(dir, 'eigen3')
            if output:
                # Only print this if the eigen3 addition was key to finding it.
                print("\n  ", dir, "  (yes)")
            return dir
        if output: print("  (no)")

    if output:
        print("Could not find Eigen in any of the standard locations.")
        print("Will now try to download it from gitlab.com. This requires an internet")
        print("connection, so it will fail if you are currently offline.")
        print("If Eigen is installed in a non-standard location, and you want to use that")
        print("instead, you should make sure the right directory is either in your")
        print("C_INCLUDE_PATH or specified in an EIGEN_DIR environment variable.")

    try:
        dir = 'downloaded_eigen'
        if os.path.isdir(dir):
            # If this exists, it was tried above and failed.  Something must be wrong with it.
            print("Previous attempt to download eigen found. Deleting and trying again.")
            shutil.rmtree(dir)
        os.mkdir(dir)
        url = 'https://gitlab.com/libeigen/eigen/-/archive/3.4.0/eigen-3.4.0.tar.bz2'
        if output:
            print("Downloading eigen from ",url)
        # Unfortunately, gitlab doesn't allow direct downloads. We need to spoof the request
        # so it thinks we're a web browser.
        # cf. https://stackoverflow.com/questions/42863240/how-to-get-round-the-http-error-403-forbidden-with-urllib-request-using-python
        page=urllib2.Request(url,headers={'User-Agent': 'Mozilla/5.0'})
        data=urllib2.urlopen(page).read()
        fname = 'eigen.tar.bz2'
        with open(fname, 'wb') as f:
            f.write(data)
        if output:
            print("Downloaded %s.  Unpacking tarball."%fname)
        with tarfile.open(fname) as tar:

            def is_within_directory(directory, target):

                abs_directory = os.path.abspath(directory)
                abs_target = os.path.abspath(target)

                prefix = os.path.commonprefix([abs_directory, abs_target])

                return prefix == abs_directory

            def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
                # Avoid security vulnerability in tar.extractall function.
                # This bit of code was added by the Advanced Research Center at Trellix in PR #1188.
                # For more information about the security vulnerability, see
                # https://github.com/advisories/GHSA-gw9q-c7gh-j9vm

                for member in tar.getmembers():
                    member_path = os.path.join(path, member.name)
                    if not is_within_directory(path, member_path):
                        raise Exception("Attempted Path Traversal in Tar File")

                tar.extractall(path, members, numeric_owner=numeric_owner)

            safe_extract(tar, dir)
        os.remove(fname)
        # This actually extracts into a subdirectory with a name eigen-eigen-5a0156e40feb/
        # I'm not sure if that name is reliable, so use glob to get it.
        dir = glob.glob(os.path.join(dir,'*'))[0]
        if os.path.isfile(os.path.join(dir, 'Eigen/Core')):
            return dir
        elif output:
            print("Downloaded eigen, but it didn't have the expected Eigen/Core file.")
    except Exception as e:
        if output:
            print("Error encountered while downloading Eigen from the internet")
            print(e)

    raise OSError("Could not find Eigen")


def try_compile(cpp_code, compiler, cflags=[], lflags=[], prepend=None, check_warning=False):
    """Check if compiling some code with the given compiler and flags works properly.
    """
    # Put the temporary files in a local tmp directory, so that they stick around after failures.
    if not os.path.exists(local_tmp): os.makedirs(local_tmp)

    # We delete these manually if successful.  Otherwise, we leave them in the tmp directory
    # so the user can troubleshoot the problem if they were expecting it to work.
    with tempfile.NamedTemporaryFile(delete=False, suffix='.cpp', dir=local_tmp) as cpp_file:
        cpp_file.write(cpp_code.encode())
        cpp_name = cpp_file.name

    # Just get a named temporary file to write to:
    with tempfile.NamedTemporaryFile(delete=False, suffix='.o', dir=local_tmp) as o_file:
        o_name = o_file.name

    # Another named temporary file for the executable
    with tempfile.NamedTemporaryFile(delete=False, suffix='.exe', dir=local_tmp) as exe_file:
        exe_name = exe_file.name

    # Try compiling with the given flags
    cc = [compiler.compiler_so[0]]
    if prepend:
        cc = [prepend] + cc
    cmd = cc + compiler.compiler_so[1:] + cflags + ['-c',cpp_name,'-o',o_name]
    if debug:
        print('cmd = ',' '.join(cmd))
    try:
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        lines = p.stdout.readlines()
        p.communicate()
        if debug and p.returncode != 0:
            print('Trying compile command:')
            print(' '.join(cmd))
            print('Output was:')
            print('   ',b'   '.join(lines).decode())
        returncode = p.returncode
        if check_warning:
            output = b' '.join(lines).decode()
            if 'warning' in output:
                returncode = 1
    except OSError as e:
        if debug:
            print('Trying compile command:')
            print(cmd)
            print('Caught error: ',repr(e))
        returncode = 1
    if returncode != 0:
        # Don't delete files in case helpful for troubleshooting.
        return False

    # Link
    cc = compiler.linker_so[0]
    cmd = [cc] + compiler.linker_so[1:] + lflags + [o_name,'-o',exe_name]
    if debug:
        print('cmd = ',' '.join(cmd))
    try:
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        lines = p.stdout.readlines()
        p.communicate()
        if debug and p.returncode != 0:
            print('Trying link command:')
            print(' '.join(cmd))
            print('Output was:')
            #print('   ',b'   '.join(lines).decode())
            print('   ',b'   '.join(lines))
        returncode = p.returncode
    except OSError as e:
        if debug:
            print('Trying link command:')
            print(' '.join(cmd))
            print('Caught error: ',repr(e))
        returncode = 1

    if returncode:
        # The linker needs to be a c++ linker, which isn't 'cc'.  However, I couldn't figure
        # out how to get setup.py to tell me the actual command to use for linking.  All the
        # executables available from build_ext.compiler.executables are 'cc', not 'c++'.
        # I think this must be related to the bugs about not handling c++ correctly.
        #    http://bugs.python.org/issue9031
        #    http://bugs.python.org/issue1222585
        # So just switch it manually and see if that works.
        if 'clang' in cc:
            cpp = cc.replace('clang', 'clang++')
        elif 'icc' in cc:
            cpp = cc.replace('icc', 'icpc')
        elif 'gcc' in cc:
            cpp = cc.replace('gcc', 'g++')
        elif ' cc' in cc:
            cpp = cc.replace(' cc', ' c++')
        elif cc == 'cc':
            cpp = 'c++'
        else:
            comp_type = get_compiler_type(compiler)
            if comp_type == 'gcc':
                cpp = 'g++'
            elif comp_type == 'clang':
                cpp = 'clang++'
            elif comp_type == 'icc':
                cpp = 'g++'
            else:
                cpp = 'c++'
        # Finally, if GALSIM_CXX is in the environment, let that take precedence.
        # (I don't know if it's safe to use a user's CXX always, so make sure the user really
        # meant to direct GalSim to use some other compiler by requiring the GALSIM prefix.)
        cpp = os.environ.get('GALSIM_CXX', cpp)
        cmd = [cpp] + compiler.linker_so[1:] + lflags + [o_name,'-o',exe_name]
        if debug:
            print('cmd = ',' '.join(cmd))
        try:
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            lines = p.stdout.readlines()
            p.communicate()
            if debug and p.returncode != 0:
                print('Trying link command:')
                print(' '.join(cmd))
                print('Output was:')
                #print('   ',b'   '.join(lines).decode())
                print('   ',b'   '.join(lines))
            returncode = p.returncode
        except OSError as e:
            if debug:
                print('Trying to link using command:')
                print(' '.join(cmd))
                print('Caught error: ',repr(e))
            returncode = 1

    # Remove the temp files
    if returncode != 0:
        # Don't delete files in case helpful for troubleshooting.
        return False
    else:
        os.remove(cpp_name)
        os.remove(o_name)
        if os.path.exists(exe_name):
            os.remove(exe_name)
        return True

def try_openmp(compiler, cc_type):
    """
    If cc --version is not helpful, the last resort is to try each compiler type and see
    if it works.
    """
    cpp_code = """
#include <iostream>
#include <vector>
#ifdef _OPENMP
#include "omp.h"
#endif

int get_max_threads() {
#ifdef _OPENMP
    return omp_get_max_threads();
#else
    return 1;
#endif
}

int main() {
    int n = 500;
    std::vector<double> x(n,0.);

#ifdef _OPENMP
#pragma omp parallel for schedule(static)
#endif
    for (int i=0; i<n; ++i) x[i] = 2*i+1;

    double sum = 0.;
    for (int i=0; i<n; ++i) sum += x[i];
    // Sum should be n^2 = 250000

    std::cout<<get_max_threads()<<"  "<<sum<<std::endl;
    return 0;
}
"""
    extra_cflags = copt[cc_type]
    extra_lflags = lopt[cc_type]
    success = try_compile(cpp_code, compiler, extra_cflags, extra_lflags)
    if not success:
        # In case libc++ doesn't work, try letting the system use the default stdlib
        try:
            extra_cflags.remove('-stdlib=libc++')
            extra_lflags.remove('-stdlib=libc++')
        except (AttributeError, ValueError):
            pass
        else:
            success = try_compile(cpp_code, compiler, extra_cflags, extra_lflags)
    return success


def try_cpp(compiler, cflags=[], lflags=[], prepend=None):
    """Check if compiling a simple bit of c++ code with the given compiler works properly.
    """
    from textwrap import dedent
    cpp_code = dedent("""
    #include <iostream>
    #include <vector>
    int main() {
        int n = 500;
        std::vector<double> x(n,0.);
        for (int i=0; i<n; ++i) x[i] = 2*i+1;
        double sum=0.;
        for (int i=0; i<n; ++i) sum += x[i];
        return sum;
    }
    """)
    return try_compile(cpp_code, compiler, cflags, lflags, prepend=prepend)

def try_cpp11(compiler, cflags=[], lflags=[], check_warning=False):
    """Check if compiling c++11 code with the given compiler works properly.
    """
    from textwrap import dedent
    cpp_code = dedent("""
    #include <iostream>
    #include <forward_list>
    #include <cmath>

    int main(void) {
        std::cout << std::tgamma(1.3) << std::endl;
        return 0;
    }
    """)
    return try_compile(cpp_code, compiler, cflags, lflags, check_warning=check_warning)


def cpu_count():
    """Get the number of cpus
    """
    try:
        import psutil
        return psutil.cpu_count()
    except ImportError:
        pass

    if hasattr(os, 'sysconf'):
        if 'SC_NPROCESSORS_ONLN' in os.sysconf_names:
            # Linux & Unix:
            ncpus = os.sysconf('SC_NPROCESSORS_ONLN')
            if isinstance(ncpus, int) and ncpus > 0:
                return ncpus
        else: # OSX:
            p = subprocess.Popen(['sysctl -n hw.ncpu'],stdout=subprocess.PIPE,shell=True)
            return int(p.stdout.read().strip())
    # Windows:
    if 'NUMBER_OF_PROCESSORS' in os.environ:
        ncpus = int(os.environ['NUMBER_OF_PROCESSORS'])
        if ncpus > 0:
            return ncpus
    return 1 # Default

def parallel_compile(self, sources, output_dir=None, macros=None,
                     include_dirs=None, debug=0, extra_preargs=None,
                     extra_postargs=None, depends=None):
    """New compile function that we monkey patch into the existing compiler instance.
    """
    import multiprocessing.pool

    # Copied from the regular compile function
    macros, objects, extra_postargs, pp_opts, build = \
            self._setup_compile(output_dir, macros, include_dirs, sources,
                                depends, extra_postargs)
    cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)

    def _single_compile(obj):
        try:
            src, ext = build[obj]
        except KeyError:
            return
        self._compile(obj, src, ext, cc_args, extra_postargs, pp_opts)

    # Set by fix_compiler
    global glob_use_njobs
    if glob_use_njobs == 1:
        # This is equivalent to regular compile function
        for obj in objects:
            _single_compile(obj)
    else:
        # Use ThreadPool, rather than Pool, since the objects are picklable.
        pool = multiprocessing.pool.ThreadPool(glob_use_njobs)
        pool.map(_single_compile, objects)
        pool.close()
        pool.join()

    # Return *all* object filenames, not just the ones we just built.
    return objects


def fix_compiler(compiler, njobs):
    # Remove any -Wstrict-prototypes in the compiler flags (since invalid for C++)
    try:
        compiler.compiler_so.remove("-Wstrict-prototypes")
    except (AttributeError, ValueError):
        pass

    # nvc++ doesn't support -Wno-unused-result
    try:
        compiler.compiler_so.remove("-Wno-unused-result")
    except (AttributeError, ValueError):
        pass

    # Figure out what compiler it will use
    comp_type = get_compiler_type(compiler, output=True)
    cc = compiler.compiler_so[0]
    already_have_ccache = False
    if cc == 'ccache':
        already_have_ccache = True
        cc = compiler.compiler_so[1]
    if cc == comp_type:
        print('Using compiler %s'%(cc))
    else:
        print('Using compiler %s, which is %s'%(cc,comp_type))

    # Make sure the compiler works with a simple c++ code
    if not try_cpp(compiler):
        # One failure mode is that sometimes there is a -B /path/to/compiler_compat
        # which can cause problems.  If we get here, try removing that.
        success = False
        if '-B' in compiler.linker_so:
            for i in range(len(compiler.linker_so)):
                if (compiler.linker_so[i] == '-B' and
                    'compiler_compat' in compiler.linker_so[i+1]):
                    print('Removing potentially problematic -B compiler_compat flags')
                    del compiler.linker_so[i:i+2]
                    success = try_cpp(compiler)
                    break
        if not success:
            print("There seems to be something wrong with the compiler or cflags")
            print(str(compiler.compiler_so))
            raise OSError("Compiler does not work for compiling C++ code")

    # Check if we can use ccache to speed up repeated compilation.
    if not already_have_ccache and try_cpp(compiler, prepend='ccache'):
        print('Using ccache')
        compiler.set_executable('compiler_so', ['ccache'] + compiler.compiler_so)

    if njobs > 1:
        # Global variable for tracking the number of jobs to use.
        # We can't pass this to parallel compile, since the signature is fixed.
        # So if using parallel compile, set this value to use within parallel compile.
        global glob_use_njobs
        glob_use_njobs = njobs
        compiler.compile = types.MethodType(parallel_compile, compiler)

    extra_cflags = copt[comp_type]
    extra_lflags = lopt[comp_type]

    success = try_cpp11(compiler, extra_cflags, extra_lflags)
    if not success:
        # In case libc++ doesn't work, try letting the system use the default stdlib
        try:
            extra_cflags.remove('-stdlib=libc++')
            extra_lflags.remove('-stdlib=libc++')
        except (AttributeError, ValueError):
            pass
        else:
            success = try_cpp11(compiler, extra_cflags, extra_lflags)
    if not success:
        print('The compiler %s with flags %s did not successfully compile C++11 code'%
              (cc, ' '.join(extra_cflags)))
        raise OSError("Compiler is not C++-11 compatible")

    # Also see if adding -msse2 works (and doesn't give a warning)
    if '-msse2' not in extra_cflags:
        extra_cflags.append('-msse2')
    if try_cpp11(compiler, extra_cflags, extra_lflags, check_warning=True):
        print('Using cflag -msse2')
    else:
        print('warning with -msse2.')
        extra_cflags.remove('-msse2')

    # If doing develop installation, it's important for the build directory to be before any
    # other directories.  Particularly ones that might have another version of GalSim installed.
    # Otherwise the wrong library can be linked, which leads to errors.
    # So, make sure that the -Lbuild/... directive happens first among any -L directives in
    # the link flags.
    linker_so = compiler.linker_so
    # Find the first -L flag among the current flags (if any)
    for i, flag in enumerate(linker_so):
        if flag.startswith('-L'):
            print('Found link: ',i,flag)
            break
    else:
        i = len(linker_so)
    # Insert -Llib for any libs that are in build directory, to make sure they are first.
    linker_so[i:i] = ['-L' + l for l in compiler.library_dirs if l.startswith('build')]
    # Copy this list back to the compiler object
    compiler.set_executable('linker_so', linker_so)

    # Return the extra cflags, since those will be added to the build step in a different place.
    print('Using extra flags ',extra_cflags)
    return extra_cflags, extra_lflags

def add_dirs(builder, output=False):
    if debug: output = True
    # We need to do most of this both for build_clib and build_ext, so separate it out here.

    # First some basic ones we always need.
    builder.include_dirs.append('include')
    builder.include_dirs.append('include/galsim')

    import numpy
    builder.include_dirs.append(numpy.get_include())

    # Look for fftw3.
    fftw_lib = find_fftw_lib(output=output)
    fftw_libpath, fftw_libname = os.path.split(fftw_lib)
    if hasattr(builder, 'library_dirs'):
        if fftw_libpath != '':
            builder.library_dirs.append(fftw_libpath)
    fftw_include = os.path.join(os.path.split(fftw_libpath)[0], 'include')
    if os.path.isfile(os.path.join(fftw_include, 'fftw3.h')):
        print('Include directory for fftw3 is ',fftw_include)
        # Usually, the fftw3.h file is in an associated include dir, but not always.
        builder.include_dirs.append(fftw_include)
    else:
        # If not, we have our own copy of fftw3.h here.
        print('Using local copy of fftw3.h')
        builder.include_dirs.append('include/fftw3')

    # Look for Eigen/Core
    eigen_dir = find_eigen_dir(output=output)
    builder.include_dirs.append(eigen_dir)

    # Finally, add pybind11's include dir
    import pybind11
    print('PyBind11 is version ',pybind11.__version__)
    print('Looking for pybind11 header files: ')
    locations = [pybind11.get_include(user=True),
                 pybind11.get_include(user=False),
                 '/usr/include',
                 '/usr/local/include',
                 None]
    for try_dir in locations:
        if try_dir is None:
            # Last time through, raise an error.
            print("Could not find pybind11 header files.")
            print("They should have been in one of the following locations:")
            for l in locations:
                if l is not None:
                    print("   ", l)
            raise OSError("Could not find PyBind11")

        print('  ',try_dir,end='')
        if os.path.isfile(os.path.join(try_dir, 'pybind11/pybind11.h')):
            print('  (yes)')
            builder.include_dirs.append(try_dir)
            break
        else:
            print('  (no)')

def parse_njobs(njobs, task=None, command=None, maxn=4):
    """Helper function to parse njobs, which may be None (use ncpu) or an int.
    Returns an int value for njobs
    """
    if njobs is None:
        njobs = cpu_count()
        if maxn != None and njobs > maxn:
            # Usually 4 is plenty.  Testing with too many jobs tends to lead to
            # memory and timeout errors.  The user can bump this up if they want.
            njobs = maxn
        if task is not None:
            if njobs == 1:
                print('Using a single process for %s.'%task)
            else:
                print('Using %d cpus for %s'%(njobs,task))
            print('To override, you may do python setup.py %s -jN'%command)
    else:
        njobs = int(njobs)
        if task is not None:
            if njobs == 1:
                print('Using a single process for %s.'%task)
            else:
                print('Using %d cpus for %s'%(njobs,task))
    return njobs

do_output = True  # Keep track of whether we used output=True in add_dirs yet.
                  # It seems that different installation methods do things in different order,
                  # but we only want to output on the first pass through add_dirs.
                  # (Unless debug = True, then also output in the second pass.)

# Make a subclass of build_ext so we can add to the -I list.
class my_build_clib(build_clib):
    user_options = build_ext.user_options + [('njobs=', 'j', "Number of jobs to use for compiling")]

    def initialize_options(self):
        build_clib.initialize_options(self)
        self.njobs = None

    def finalize_options(self):
        global do_output
        build_clib.finalize_options(self)
        if self.njobs is None and 'glob_njobs' in globals():
            global glob_njobs
            self.njobs = glob_njobs
        add_dirs(self, output=do_output)
        do_output = False

    # Add any extra things based on the compiler being used..
    def build_libraries(self, libraries):

        build_ext = self.distribution.get_command_obj('build_ext')
        njobs = parse_njobs(self.njobs, 'compiling', 'install')

        cflags, lflags = fix_compiler(self.compiler, njobs)

        # Add the appropriate extra flags for that compiler.
        for (lib_name, build_info) in libraries:
            build_info['cflags'] = build_info.get('cflags',[]) + cflags
            build_info['lflags'] = build_info.get('lflags',[]) + lflags

        self.do_build_libraries(libraries)

    def do_build_libraries(self, libraries):
        # This version just calls the setuptools build_libraries function.
        # We'll change this for build_shared_clib below.
        build_clib.build_libraries(self, libraries)

class my_build_shared_clib(my_build_clib):

    def do_build_libraries(self, libraries):
        from distutils.errors import DistutilsSetupError
        from distutils import log
        try:
            from setuptools.modified import newer_pairwise_group
        except ImportError:
            from setuptools.dep_util import newer_pairwise_group
        from distutils.ccompiler import CCompiler

        builder = self.distribution.get_command_obj('build_ext')
        cflags, lflags = fix_compiler(self.compiler, 1)

        # Most of this is the setuptools version of the build_libraries function.
        # We just change the final link command to build a shared library that can be linked
        # to C++ programs that just want the C++ library.
        for (lib_name, build_info) in libraries:
            sources = build_info.get('sources')
            if sources is None or not isinstance(sources, (list, tuple)):
                raise DistutilsSetupError(
                       "in 'libraries' option (library '%s'), "
                       "'sources' must be present and must be "
                       "a list of source filenames" % lib_name)
            sources = list(sources)

            log.info("building '%s' library", lib_name)

            # Make sure everything is the correct type.
            # obj_deps should be a dictionary of keys as sources
            # and a list/tuple of files that are its dependencies.
            obj_deps = build_info.get('obj_deps', dict())
            if not isinstance(obj_deps, dict):
                raise DistutilsSetupError(
                    "in 'libraries' option (library '%s'), "
                    "'obj_deps' must be a dictionary of "
                    "type 'source: list'" % lib_name)
            dependencies = []

            # Get the global dependencies that are specified by the '' key.
            # These will go into every source's dependency list.
            global_deps = obj_deps.get('', list())
            if not isinstance(global_deps, (list, tuple)):
                raise DistutilsSetupError(
                    "in 'libraries' option (library '%s'), "
                    "'obj_deps' must be a dictionary of "
                    "type 'source: list'" % lib_name)

            # Build the list to be used by newer_pairwise_group
            # each source will be auto-added to its dependencies.
            for source in sources:
                src_deps = [source]
                src_deps.extend(global_deps)
                extra_deps = obj_deps.get(source, list())
                if not isinstance(extra_deps, (list, tuple)):
                    raise DistutilsSetupError(
                        "in 'libraries' option (library '%s'), "
                        "'obj_deps' must be a dictionary of "
                        "type 'source: list'" % lib_name)
                src_deps.extend(extra_deps)
                dependencies.append(src_deps)

            expected_objects = self.compiler.object_filenames(
                sources,
                output_dir=self.build_temp,
            )

            if (
                newer_pairwise_group(dependencies, expected_objects)
                != ([], [])
            ):
                # First, compile the source code to object files in the library
                # directory.  (This should probably change to putting object
                # files in a temporary build directory.)
                macros = build_info.get('macros')
                include_dirs = build_info.get('include_dirs')
                cflags = build_info.get('cflags')
                self.compiler.compile(
                    sources,
                    output_dir=self.build_temp,
                    macros=macros,
                    include_dirs=include_dirs,
                    extra_postargs=cflags,
                    debug=self.debug
                )

            ###
            ###
            ### This is the one bit that is changed from the setuptools version.
            ### The original used self.compiler.create_static_lib
            ###
            ###
            lib_name = self.compiler.library_filename(lib_name, lib_type="shared")
            version_str = '{}.{}'.format(*version_info[:2])
            if sys.platform == "darwin":
                # .so -> .dylib
                lib_name = lib_name.replace('so','dylib')
                full_lib_name = lib_name[:-5] + version_str + '.dylib'
                orig_linker_so = self.compiler.linker_so
                assert orig_linker_so[1] == '-bundle'
                dylib_linker_so = orig_linker_so.copy()
                dylib_linker_so[1] = '-dynamiclib'
                self.compiler.set_executable('linker_so', dylib_linker_so)
            else:
                # Just add the version bit
                full_lib_name = lib_name[:-2] + version_str + '.so'

            libraries = builder.get_libraries(ext)
            print('initial libraries = ',libraries)

            library_dirs = ext.library_dirs
            print('initial library_dirs = ',library_dirs)

            # Link to fftw library
            fftw_lib = find_fftw_lib()
            fftw_libpath, fftw_libname = os.path.split(fftw_lib)
            if fftw_libpath != '':
                library_dirs.append(fftw_libpath)
            libraries.append(fftw_libname.split('.')[0][3:])

            # Check for conda libraries that might host OpenMP
            env = dict(os.environ)
            if 'CONDA_PREFIX' in env:
                library_dirs.append(env['CONDA_PREFIX']+'/lib')

            if sys.platform == 'darwin':
                # Set the compatibility version on macos
                lflags.extend(['-Wl,-compatibility_version,%s.%s'%version_info[:2],
                               '-Wl,-current_version,%s.%s.%s'%version_info ])
                # Also add rpath specification for fftw
                if fftw_libpath != '':
                    lflags.append('-Wl,-rpath,' + fftw_libpath)
                # Set the install_name to the right rpath name.
                lflags.append('-Wl,-install_name,@rpath/%s'%full_lib_name)

            output_dir = os.path.join('build','shared_clib')
            self.compiler.link(CCompiler.SHARED_OBJECT, expected_objects, full_lib_name,
                               libraries=libraries,
                               library_dirs=library_dirs,
                               runtime_library_dirs=ext.runtime_library_dirs,
                               extra_postargs=ext.extra_link_args + lflags,
                               output_dir=output_dir, debug=self.debug)

            # Also make the non-versionful one
            full_lib_name_with_dir = os.path.join(output_dir, full_lib_name)
            lib_name_with_dir = os.path.join(output_dir, lib_name)
            print('Versioned library: ',full_lib_name_with_dir)
            print('Un-versioned library: ',lib_name_with_dir)
            if not os.path.exists(lib_name_with_dir):
                # This is slightly confusing.
                # The target needs the dir, but the source cannot include the dir or it
                # will try to link to build/shared_clib/build/shared_clib/libgalsim...
                os.symlink(full_lib_name, lib_name_with_dir)



# Make a subclass of build_ext so we can add to the -I list.
class my_build_ext(build_ext):
    user_options = build_ext.user_options + [('njobs=', 'j', "Number of jobs to use for compiling")]

    def initialize_options(self):
        build_ext.initialize_options(self)
        self.njobs = None

    def finalize_options(self):
        global do_output
        build_ext.finalize_options(self)
        # I couldn't find an easy way to send the user option from my_install to my_buld_ext.
        # So use a global variable. (UGH!)
        if self.njobs is None and 'glob_njobs' in globals():
            global glob_njobs
            self.njobs = glob_njobs
        add_dirs(self, output=do_output)
        do_output = False

    # Add any extra things based on the compiler being used..
    def build_extensions(self):

        njobs = parse_njobs(self.njobs, 'compiling', 'install')
        cflags, lflags = fix_compiler(self.compiler, njobs)

        # Add the appropriate extra flags for that compiler.
        for e in self.extensions:
            e.extra_compile_args = cflags
            for flag in lflags:
                e.extra_link_args.append(flag)

        # Now run the normal build function.
        build_ext.build_extensions(self)

    def run(self):
        # `python setup.py develop` (used implicitly by `pip install -e .`) only calls
        # `build_ext`, unlike `install` which calls `build` and all its related
        # sub-commands. Linking -lgalsim fails, since `build_clib` isn't calld.
        # This override ensures that `build_clib` command is run before the
        # pysrc stuff is linked.
        # cf. https://github.com/pypa/pip/issues/4523

        self.run_command("build_clib")
        build_ext.run(self)

        # If requested, also build the shared library.
        if int(os.environ.get('GALSIM_BUILD_SHARED', 0)):
            self.run_command("build_shared_clib")

        if int(os.environ.get('GALSIM_RUN_TEST', 0)):
            self.run_command("run_cpp_test")


class my_install(install):
    user_options = install.user_options + [('njobs=', 'j', "Number of jobs to use for compiling")]

    def initialize_options(self):
        install.initialize_options(self)
        self.njobs = None

    def finalize_options(self):
        install.finalize_options(self)
        global glob_njobs
        glob_njobs = self.njobs


# AFAICT, setuptools doesn't provide any easy access to the final installation location of the
# executable scripts.  This bit is just to save the value of script_dir so I can use it later.
# cf. http://stackoverflow.com/questions/12975540/correct-way-to-find-scripts-directory-from-setup-py-in-python-distutils/
class my_easy_install(easy_install):    # Used when installing via python setup.py install
    # Match the call signature of the easy_install version.
    def write_script(self, script_name, contents, mode="t", *ignored):
        # Run the normal version
        easy_install.write_script(self, script_name, contents, mode, *ignored)
        # Save the script install directory in the distribution object.
        # This is the same thing that is returned by the setup function.
        self.distribution.script_install_dir = self.script_dir

class my_install_scripts(install_scripts):  # Used when pip installing.
    def run(self):
        install_scripts.run(self)
        self.distribution.script_install_dir = self.install_dir

class my_run_cpp_test(my_build_ext):

    def run(self):
        builder = self.distribution.get_command_obj('build_ext')
        compiler = builder.compiler
        cflags, lflags = fix_compiler(compiler, 1)

        ext = builder.extensions[0]
        objects = compiler.compile(test_sources,
                output_dir=builder.build_temp,
                macros=ext.define_macros,
                include_dirs=ext.include_dirs,
                extra_postargs=cflags,
                debug=builder.debug,
                depends=ext.depends)

        if ext.extra_objects:
            objects.extend(ext.extra_objects)

        libraries = builder.get_libraries(ext)
        libraries.append('galsim')
        library_dirs = ext.library_dirs

        # Use the shared library when building the c++ executables to make sure it works.
        self.run_command("build_shared_clib")
        library_dirs.append('build/shared_clib')

        # Check for conda libraries that might host OpenMP
        env = dict(os.environ)
        if 'CONDA_PREFIX' in env:
            library_dirs.append(env['CONDA_PREFIX']+'/lib')

        exe_file = os.path.join(builder.build_temp,'cpp_test')
        compiler.link_executable(
                objects, 'cpp_test',
                output_dir=builder.build_temp,
                libraries=libraries,
                library_dirs=library_dirs,
                runtime_library_dirs=ext.runtime_library_dirs,
                debug=builder.debug,
                target_lang='c++')

        # Might need extra dirs in LD_LIBRARY_PATH.  Just add them to make sure.
        for flag in compiler.linker_so:
            if flag.startswith('-L'):
                library_dirs.append(flag[2:])
        if 'LD_LIBRARY_PATH' not in env:
            env['LD_LIBRARY_PATH'] = env.get('LSST_LIBRARY_PATH','')
        if 'DYLD_LIBRARY_PATH' not in env:
            env['DYLD_LIBRARY_PATH'] = env.get('LSST_LIBRARY_PATH','')
        env['LD_LIBRARY_PATH'] += ':' + ':'.join(library_dirs)
        env['DYLD_LIBRARY_PATH'] += ':' + ':'.join(library_dirs)

        # Run the test executable.
        # And pass this env to the execution environment.
        p = subprocess.Popen([exe_file], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env)
        lines = p.stdout.readlines()
        p.communicate()
        for line in lines:
            print(line.decode().strip())
        if p.returncode != 0:
            raise RuntimeError("C++ tests failed")
        print("All C++ tests passed.")


lib=("galsim", {'sources' : cpp_sources,
                'depends' : headers + inst,
                'include_dirs' : ['include', 'include/galsim'],
                'undef_macros' : undef_macros })
ext=Extension("galsim._galsim",
              py_sources,
              depends = cpp_sources + headers + inst,
              undef_macros = undef_macros,
              extra_link_args = ["-lfftw3"])

build_dep = ['setuptools>=38', 'pybind11>=2.2', 'numpy>=1.17']
run_dep = ['astropy', 'LSSTDESC.Coord']
test_dep = ['pytest', 'pytest-xdist', 'pytest-timeout', 'scipy', 'pyyaml']

# If Eigen doesn't exist in the normal places, download it.
find_eigen_dir(output=True)

with open('README.rst') as file:
    long_description = file.read()

# Read in the galsim version from galsim/_version.py
# cf. http://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
version_file=os.path.join('galsim','_version.py')
verstrline = open(version_file, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    galsim_version = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (version_file,))
print('GalSim version is %s'%(galsim_version))

# Write a Version.h file that has this information for people using the C++ library.
vi = re.split(r'\.|-',galsim_version)
version_info = tuple([int(x) for x in vi if x.isdigit()])
if len(version_info) == 2:
    version_info = version_info + (0,)
version_h_text = """
// This file is auto-generated by setup.py.  Do not edit.
#define GALSIM_MAJOR %d
#define GALSIM_MINOR %d
#define GALSIM_REVISION %d

#include <string>
#include <sstream>

#if defined(__GNUC__)
#define PUBLIC_API __attribute__ ((visibility ("default")))
#else
#define PUBLIC_API
#endif

namespace galsim {
    // Compiled versions of the above #define values.
    PUBLIC_API int major_version();
    PUBLIC_API int minor_version();
    PUBLIC_API int revision();

    // Returns string of the form "1.4.2"
    PUBLIC_API std::string version();

    // Checks if the compiled library version matches the #define values in this header file.
    inline bool check_version() {
        // Same code as version(), but inline, so we get the above values to compare
        // to the values compiled into the library.
        std::ostringstream oss;
        oss << GALSIM_MAJOR << '.' << GALSIM_MINOR << '.' << GALSIM_REVISION;
        return oss.str() == version();
    }
}
"""%version_info[:3]
version_h_file = os.path.join('include', 'galsim', 'Version.h')
with open(version_h_file, 'w') as f:
    f.write(version_h_text)

headers.append(version_h_file)

dist = setup(name="GalSim",
    version=galsim_version,
    author="GalSim Developers (point of contact: Mike Jarvis)",
    author_email="michael@jarvis.net",
    description="The modular galaxy image simulation toolkit",
    long_description=long_description,
    license = "BSD License",
    url="https://github.com/GalSim-developers/GalSim",
    download_url="https://github.com/GalSim-developers/GalSim/releases/tag/v%s.zip"%galsim_version,
    packages=find_packages(),
    package_data={'galsim' : shared_data + headers},
    libraries=[lib],
    ext_modules=[ext],
    setup_requires=build_dep,
    install_requires=build_dep + run_dep,
    tests_require=test_dep,
    cmdclass = {'build_ext': my_build_ext,
                'build_clib': my_build_clib,
                'build_shared_clib': my_build_shared_clib,
                'install': my_install,
                'install_scripts': my_install_scripts,
                'easy_install': my_easy_install,
                'run_cpp_test': my_run_cpp_test,
                },
    entry_points = {'console_scripts' : [
            'galsim = galsim.__main__:run_main',
            'galsim_download_cosmos = galsim.download_cosmos:run_main'
            ]},
    zip_safe=False,
    )

# Check that the path includes the directory where the scripts are installed.
real_env_path = [os.path.realpath(d) for d in os.environ['PATH'].split(':')]
if hasattr(dist,'script_install_dir'):
    print('scripts installed into ',dist.script_install_dir)
    if (dist.script_install_dir not in os.environ['PATH'].split(':') and
        os.path.realpath(dist.script_install_dir) not in real_env_path):

        print('\nWARNING: The GalSim executables were installed in a directory not in your PATH')
        print('         If you want to use the executables, you should add the directory')
        print('\n             ',dist.script_install_dir,'\n')
        print('         to your path.  The current path is')
        print('\n             ',os.environ['PATH'],'\n')
        print('         Alternatively, you can specify a different prefix with --prefix=PREFIX,')
        print('         in which case the scripts will be installed in PREFIX/bin.')
        print('         If you are installing via pip use --install-option="--prefix=PREFIX"')



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import os
import logging

from .output import OutputBuilder, RegisterOutputType
from .image import BuildImages
from .input import ProcessInputNObjects
from .value import ParseValue, CheckAllParams
from ..errors import GalSimConfigError

class MultiFitsBuilder(OutputBuilder):
    """Builder class for constructing and writing MultiFits output types.
    """

    def buildImages(self, config, base, file_num, image_num, obj_num, ignore, logger):
        """Build the images

        Parameters:
            config:         The configuration dict for the output field.
            base:           The base configuration dict.
            file_num:       The current file_num.
            image_num:      The current image_num.
            obj_num:        The current obj_num.
            ignore:         A list of parameters that are allowed to be in config that we can
                            ignore here.  i.e. it won't be an error if they are present.
            logger:         If given, a logger object to log progress.

        Returns:
            a list of the images built
        """
        nimages = self.getNImages(config, base, file_num, logger=logger)

        # The above call sets up a default nimages if appropriate.  Now, check that there are no
        # invalid parameters in the config dict.
        req = { 'nimages' : int }
        ignore += [ 'file_name', 'dir', 'nfiles' ]
        CheckAllParams(config, ignore=ignore, req=req)

        return BuildImages(nimages, base, image_num, obj_num, logger=logger)

    def getNImages(self, config, base, file_num, logger=None):
        """
        Get the number of images for a MultiFits file type.

        Parameters:
            config:         The configuration dict for the output field.
            base:           The base configuration dict.
            file_num:       The current file number.
            logger:         If given, a logger object to log progress.

        Returns:
            the number of images
        """
        # Allow nimages to be automatic based on input catalog if image type is Single
        if ( 'nimages' not in config and
            ( 'image' not in base or 'type' not in base['image'] or
            base['image']['type'] == 'Single' ) ):
            nimages = ProcessInputNObjects(base)
            if nimages:
                config['nimages'] = nimages
        if 'nimages' not in config:
            raise GalSimConfigError(
                "Attribute output.nimages is required for output.type = MultiFits")
        return ParseValue(config,'nimages',base,int)[0]


# Register this as a valid output type
RegisterOutputType('MultiFits', MultiFitsBuilder())



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#
"""
Module containing code for estimating shear power spectra from shears at gridded positions.

The code below was developed largely by Joe Zuntz and tweaked by assorted GalSim
developers.  This development and testing took place in a separate (private) repository before the
code was moved into the GalSim repository, but there are some demonstrations of the performance of
this code in devel/modules/lensing_engine.pdf
"""
import numpy as np
import os
import sys

from .errors import GalSimError, GalSimValueError, GalSimIncompatibleValuesError
from .table import LookupTable


class PowerSpectrumEstimator:
    """Class for estimating the shear power spectrum from gridded shears.

    This class stores all the data used in power spectrum estimation that is fixed with the geometry
    of the problem - the binning and spin weighting factors.

    The only public method is estimate(), which is called with 2D ``g1`` and ``g2`` arrays on a
    square grid.  It assumes the flat sky approximation (where ``ell`` and ``k`` are
    interchangeable), and rebins the observed ell modes into a user-defined number of logarithimic
    bins in ell.  Given that the grid parameters are precomputed and stored when the
    `PowerSpectrumEstimator` is initialized, computation of the PS for multiple sets of shears
    corresponding to the same grid setup can proceed more rapidly than if everything had to be
    recomputed each time.

    Below is an example of how to use this code (relying on GalSim to provide the arrays of g1 and
    g2, though that is by no means required, and assuming that the user is sitting in the examples/
    directory)::

        >>> grid_size = 10.  # Define the total grid extent, in degrees
        >>> ngrid = 100      # Define the number of grid points in each dimension: (ngrid x ngrid)
        >>> n_ell = 15       # Choose the number of logarithmic bins in ell or k for outputs
        >>>
        >>> # Define a lookup-table for the power spectrum as a function of k based on the outputs
        >>> # of iCosmo (see demo11.py for more description of how this was generated).
        >>> my_tab = galsim.LookupTable(file='data/cosmo-fid.zmed1.00.out')
        >>>
        >>> # Generate a galsim.PowerSpectrum with this P(k), noting the units.
        >>> my_ps = galsim.PowerSpectrum(my_tab, units=galsim.radians)
        >>>
        >>> # Build a grid of shear values with the desired parameters.
        >>> g1, g2 = my_ps.buildGrid(grid_spacing=grid_size/ngrid, ngrid=ngrid,
        ...                          units=galsim.degrees)
        >>>
        >>> # Initialize a PowerSpectrumEstimator with the chosen grid geometry and number of ell
        >>> # bins. Note that these values are actually the default, so we didn't technically have
        >>> # to specifythem.
        >>> my_pse = galsim.pse.PowerSpectrumEstimator(ngrid, grid_size, n_ell)
        >>>
        >>> # Estimate the power based on this set of g1, g2.  If we get another set of shears for
        >>> # the same grid geometry, we can reuse the same PowerSpectrumEstimator object.
        >>> ell, P_e, P_b, P_eb = my_pse.estimate(g1, g2)

    The output NumPy arrays ``ell``, ``P_e``, ``P_b``, and ``P_eb`` contain the effective ell
    value, the E-mode auto-power spectrum, the B-mode auto-power spectrum, and the EB cross-power
    spectrum.  The units are inverse radians for ell, and radians^2 for the output power spectra.

    Some important notes:

    1) Power spectrum estimation requires a weight function which decides how the averaging is done
       across ell within each bin.  By default that weighting is flat in ell using an analytic
       calculation of the area in ell space, but this is easy to change with the ``_bin_power``
       function.  (Note this area averaged bin weighting is only approximate for the higher
       frequency bins in which the lower ``ell`` edge is greater than ``pi * ngrid / grid_size``,
       due to the annular ``ell`` region being cut off by the square grid edges beyond this value.)
       A keyword allows for weighting by the power itself.
    2) This is the power spectrum of the gridded *data*, not the underlying field - we do not
       account for the effects of the finite grid (basically, ignoring all the reasons why power
       spectrum estimation is hard - see devel/modules/lensing_engine.pdf in the GalSim repository).
       Users must account for the contribution of noise in ``g1``, ``g2`` and any masking.
    3) The binning is currently fixed as uniform in log(ell).
    4) The code for this class uses the notation of the GREAT10 handbook (Kitching et al. 2011,
       http://dx.doi.org/10.1214/11-AOAS484), equations 17-21.
    """
    def __init__(self, N=100, sky_size_deg=10., nbin=15):
        """Create a PowerSpectrumEstimator object given some grid parameters.

        This constructor precomputes some numbers related to the grid geometry, so the same
        PowerSpectrumEstimator can be used to estimate the power spectrum quickly for many sets of
        shears at gridded positions.

        Parameters:
            N:              The number of pixels along each side of the grid. [default: 100]
            sky_size_deg:   The total grid width (in one dimension) in degrees. [default: 10]
            nbin:           The number of evenly-spaced logarithmic ``ell`` bins to use for
                            estimating the power spectrum. [default: 15]
        """
        # Set up the scales of the sky and pixels
        self.N = N
        self.sky_size_deg = sky_size_deg
        self.nbin = nbin
        self.sky_size = np.radians(sky_size_deg)
        self.dx = self.sky_size / N

        # Define the possible ell range, the bin edges and effective ell values.
        # This is necessary for binning the power spectrum in ell.
        lmin = 2*np.pi / self.sky_size
        lmax = np.sqrt(2.)*np.pi / self.dx # in 2 dimensions
        self.bin_edges = np.logspace(np.log10(lmin), np.log10(lmax), nbin+1)
        # By default, report an area-averaged value of ell, which should be fine if there is
        # no weighting (in which case it's recomputed) and if there are many ell modes in
        # each bin.  The latter assumption is most likely to break down at low ell.  Note also that
        # at high ell when the lower ell edge is greater than pi * ngrid / grid_size, due to the
        # annular ell region being cut off by the square grid edges beyond this value, this annular
        # average is only approximate.
        self.ell = (2./3.)*(self.bin_edges[1:]**3-self.bin_edges[:-1]**3) \
                                   / (self.bin_edges[1:]**2-self.bin_edges[:-1]**2)

        # Precompute and store two useful factors, both in the form of 2D grids in Fourier space.
        # These are the lengths of the wavevector |ell| for each point in the space, and the complex
        # valued spin-weighting that takes the complex shear fields -> E,B
        self.l_abs, self.eb_rot = self._generate_eb_rotation()

    def __repr__(self):
        return "galsim.pse.PowerSpectrumEstimator(N=%r, sky_size_deg=%r, nbin=%r)"%(
                self.N, self.sky_size_deg, self.nbin)
    def __eq__(self, other): return self is other or repr(self) == repr(other)
    def __ne__(self, other): return not self.__eq__(other)
    def __hash__(self): return hash(repr(self))

    def _generate_eb_rotation(self):
        # Set up the Fourier space grid lx, ly.
        ell = 2*np.pi*np.fft.fftfreq(self.N, self.dx)
        lx, ly = np.meshgrid(ell,ell)

        # Now compute the lengths and angles of the ell vectors.
        l_sq = lx**2 + ly**2

        # Compute exp(-2i psi) where psi = atan2(ly,lx)
        l_sq[0,0] = 1  # Avoid division by 0
        expm2ipsi = (lx - 1j * ly)**2 / l_sq
        l_abs = np.sqrt(l_sq)
        l_abs[0,0] = 0  # Go back to correct value at 0,0.

        self.lx = lx
        self.ly = ly

        return l_abs, expm2ipsi

    def _bin_power(self, C, ell_weight=None):
        # This little utility function bins a 2D C^{E/B, E/B}_{ell} based on |ell|.  The use of
        # histogram is a little hack, but is quite convenient since it means everything is done in C
        # so it is very fast. The first call to `histogram` just returns an array over the
        # logarithmic ell bins of
        # sum_{|ell| in bin} weight(|ell|)*C_{ell_x,ell_y}
        # and the second call returns
        # sum_{|ell| in bin} weight(|ell|).
        # Thus, the ratio is just the mean power in the bin.  If `ell_weight` is None, then weight=1
        # for all ell, corresponding to a simple averaging process.  If `ell_weight` is not None,
        # then some non-flat weighting scheme is used for averaging over the ell values within a
        # bin.
        if ell_weight is not None:
            ell_weight = np.abs(ell_weight)
            P,_ = np.histogram(self.l_abs, self.bin_edges, weights=C*ell_weight)
            count,_ = np.histogram(self.l_abs, self.bin_edges, weights=ell_weight)
        else:
            P,_ = np.histogram(self.l_abs, self.bin_edges, weights=C)
            count,_ = np.histogram(self.l_abs, self.bin_edges)
        if (count == 0).any():
            raise GalSimError("Logarithmic bin definition resulted in >=1 empty bin!")
        return P/count

    def estimate(self, g1, g2, weight_EE=False, weight_BB=False, theory_func=None):
        """Compute the EE, BB, and EB power spectra of two 2D arrays ``g1`` and ``g2``.

        For example usage, see the docstring for the `PowerSpectrumEstimator` class.

        Parameters:
            g1:             The shear component g1 as a square 2D NumPy array.
            g2:             The shear component g2 as a square 2D NumPy array.
            weight_EE:      If True, then the E auto-power spectrum is re-computed weighting by
                            the power within each logarithmically-spaced ell bin. [default: False]
            weight_BB:      If True, then the B auto-power spectrum is re-computed weighting by
                            the power within each logarithmically-spaced ell bin. [default: False]
            theory_func:    An optional callable function that can be used to get an idealized
                            value of power at each point on the grid, and then see what results
                            it gives for our chosen ell binning. [default: None]
        """
        # Check for the expected square geometry consistent with the previously-defined grid size.
        if g1.shape != g2.shape:
            raise GalSimIncompatibleValuesError(
                "g1 and g2 grids do not have the same shape.", g1=g1, g2=g2)
        if g1.shape[0] != g1.shape[1]:
            raise GalSimValueError("Input shear arrays must be square.", g1.shape)
        if g1.shape[0] != self.N:
            raise GalSimValueError("Input shear array size is not correct!", g1.shape)

        if not isinstance(weight_EE, bool) or not isinstance(weight_BB, bool):
            raise TypeError("Input weight flags must be bools!")

        # Transform g1+j*g2 into Fourier space and rotate into E-B, then separate into E and B.
        EB = np.fft.ifft2(self.eb_rot * np.fft.fft2(g1 + 1j*g2))
        E = np.fft.fft2(EB.real)
        B = np.fft.fft2(EB.imag)

        # Use the internal function above to bin, and account for the normalization of the FFT.
        # Recall that power has units of angle^2, which is the reason why we need a self.dx^2 in the
        # equations below in addition to the standard 1/N^2 coming from the FFTs.
        C_EE = self._bin_power(E*np.conjugate(E))*(self.dx/self.N)**2
        C_BB = self._bin_power(B*np.conjugate(B))*(self.dx/self.N)**2
        C_EB = self._bin_power(E*np.conjugate(B))*(self.dx/self.N)**2

        if theory_func is not None:
            # theory_func needs to be a callable function
            C_theory_ell = np.zeros_like(self.l_abs)
            C_theory_ell[self.l_abs>0] = theory_func(self.l_abs[self.l_abs>0])
            C_theory = self._bin_power(C_theory_ell)

        if weight_EE or weight_BB:
            # Need to interpolate C_EE to values of self.l_abs.  A bit of kludginess as we go off
            # the end of our final ell grid...
            new_ell = np.zeros(len(self.ell)+2)
            new_ell[1:len(self.ell)+1] = self.ell
            new_ell[len(self.ell)+1] = 10.*max(self.ell)
            if theory_func is not None:
                C_theory = self._bin_power(C_theory_ell, ell_weight=C_theory_ell)

        if weight_EE:
            new_CEE = np.zeros_like(new_ell)
            new_CEE[1:len(self.ell)+1] = np.real(C_EE)
            new_CEE[len(self.ell)+1] = new_CEE[len(self.ell)]
            EE_table = LookupTable(new_ell, new_CEE)
            ell_weight = EE_table(self.l_abs)
            C_EE = self._bin_power(E*np.conjugate(E), ell_weight=ell_weight)*(self.dx/self.N)**2

        if weight_BB:
            new_CBB = np.zeros_like(new_ell)
            new_CBB[1:len(self.ell)+1] = np.real(C_BB)
            new_CBB[len(self.ell)+1] = new_CBB[len(self.ell)]
            BB_table = LookupTable(new_ell, new_CBB)
            ell_weight = BB_table(self.l_abs)
            C_BB = self._bin_power(B*np.conjugate(B), ell_weight=ell_weight)*(self.dx/self.N)**2

        # For convenience, return ell (copied in case the user messes with it) and the three power
        # spectra. If the user requested a binned theoretical spectrum, return that as well.
        if theory_func is None:
            return self.ell.copy(), np.real(C_EE), np.real(C_BB), np.real(C_EB)
        else:
            return self.ell.copy(), np.real(C_EE), np.real(C_BB), np.real(C_EB), np.real(C_theory)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'LookupTable', 'LookupTable2D', '_LookupTable', '_LookupTable2D', 'trapz', ]

import numpy as np
import numbers
import warnings

from . import _galsim
from ._utilities import lazy_property, basestring
from .bounds import BoundsD
from .position import _PositionD
from .errors import *
from .interpolant import convert_interpolant

def _str_array(a):
    # Used by both LookupTable.__str__ and LookupTable2D.__str__
    # to write the x, f, etc. numpy arrays in an abbreviated form so they don't fill the
    # screen with numbers.
    # 1. Write the whole array if it has at most 5 values,
    # 2. Just write the first two and last two values in the array if longer.
    # 3. linewidth defaults to 75, which adds annoying linebreaks here.
    #    1000 should be big enough to mean "never".
    with np.printoptions(threshold=5, edgeitems=2, linewidth=1000):
        return repr(a)

class LookupTable:
    """
    LookupTable represents a lookup table to store function values that may be slow to calculate,
    for which interpolating from a lookup table is sufficiently accurate.

    A LookupTable may be constructed from two arrays (lists, tuples, or NumPy arrays of
    floats/doubles)::

        >>> args = [...]
        >>> vals = []
        >>> for arg in args:
        ...     val = calculateVal(arg)
        ...     vals.append(val)
        >>> table = galsim.LookupTable(x=args,f=vals)

    Then you can use this table as a replacement for the slow calculation::

        >>> other_args = [...]
        >>> for arg in other_args:
        ...     val = table(arg)
        ...     [... use val ...]


    The default interpolation method is a natural cubic spline.  This is usually the best choice,
    but we also provide other options, which can be specified by the ``interpolant`` kwarg.  The
    choices include 'floor', 'ceil', 'linear', 'spline', or a `galsim.Interpolant` object:

    - 'floor' takes the value from the previous argument in the table.
    - 'ceil' takes the value from the next argument in the table.
    - 'nearest' takes the value from the nearest argument in the table.
    - 'linear' does linear interpolation between these two values.
    - 'spline' uses a cubic spline interpolation, so the interpolated values are smooth at
      each argument in the table.
    - a `galsim.Interpolant` object or a string convertible to one.  This option can be used for
      `Lanczos` or `Quintic` interpolation, for example.

    Note that specifying the string 'nearest' or 'linear' will use a LookupTable-optimized
    interpolant instead of `galsim.Nearest` or `galsim.Linear`, though the latter options can still
    be used by passing an `Interpolant` object instead of a string.  Also note that to use a
    `galsim.Interpolant` in a LookupTable, the input data must be equally spaced, or logarithmically
    spaced if ``x_log`` is set to True (see below).  Finally, although natural cubic spline used
    when interpolant='spline' and the cubic convolution interpolant used when the interpolant
    is `galsim.Cubic` both produce piecewise cubic polynomial interpolations, their treatments of
    the continuity of derivatives are different (the natural spline is smoother).

    There are also two factory functions which can be used to build a LookupTable:

        `LookupTable.from_func`
                makes a LookupTable from a callable function

        `LookupTable.from_file`
                reads in a file of x and f values.

    The user can also opt to interpolate in log(x) and/or log(f) (if not using a
    `galsim.Interpolant`), though this is not the default.  It may be a wise choice depending on the
    particular function, e.g., for a nearly power-law f(x) (or at least one that is locally
    power-law-ish for much of the x range) then it might be a good idea to interpolate in log(x) and
    log(f) rather than x and f.

    Parameters:
        x:              The list, tuple, or NumPy array of ``x`` values.
        f:              The list, tuple, or NumPy array of ``f(x)`` values.
        interpolant:    Type of interpolation to use, with the options being 'floor', 'ceil',
                        'nearest', 'linear', 'spline', or a `galsim.Interpolant` or string
                        convertible to one.  [default: 'spline']
        x_log:          Set to True if you wish to interpolate using log(x) rather than x.  Note
                        that all inputs / outputs will still be x, it's just a question of how the
                        interpolation is done. [default: False]
        f_log:          Set to True if you wish to interpolate using log(f) rather than f.  Note
                        that all inputs / outputs will still be f, it's just a question of how the
                        interpolation is done. [default: False]
    """
    def __init__(self, x, f, interpolant='spline', x_log=False, f_log=False):
        self.x_log = x_log
        self.f_log = f_log

        # Check if interpolant is a string that we understand.  If not, try convert_interpolant
        if interpolant in ('nearest', 'linear', 'ceil', 'floor', 'spline'):
            self._interp1d = None
        else:
            self._interp1d = convert_interpolant(interpolant)
        self.interpolant = interpolant

        # Sanity checks
        if len(x) != len(f):
            raise GalSimIncompatibleValuesError("Input array lengths don't match", x=x, f=f)
        if len(x) < 2:
            raise GalSimValueError("Input arrays too small to interpolate", x)

        # turn x and f into numpy arrays so that all subsequent math is possible (unlike for
        # lists, tuples).  Also make sure the dtype is float
        x = np.asarray(x, dtype=float)
        if np.all(x[1:] >= x[:-1]):
            # Already sorted (a common case, so avoid the sort.
            self.x = np.ascontiguousarray(x, dtype=float)
            self.f = np.ascontiguousarray(f, dtype=float)
        else:
            f = np.asarray(f, dtype=float)
            s = np.argsort(x)
            self.x = np.ascontiguousarray(x[s])
            self.f = np.ascontiguousarray(f[s])

        self._x_min = self.x[0]
        self._x_max = self.x[-1]
        if self._x_min == self._x_max:
            raise GalSimValueError("All x values are equal", x)
        if self.x_log and self.x[0] <= 0.:
            raise GalSimValueError("Cannot interpolate in log(x) when table contains x<=0.", x)
        if self.f_log and np.any(self.f <= 0.):
            raise GalSimValueError("Cannot interpolate in log(f) when table contains f<=0.", f)

        # inf causes problems in some cases, so avoid it if used as the maximum x value.
        if self._x_max == np.inf:
            self.x[-1] = self._x_max = 1.e300

        # Check equal-spaced arrays
        if self._interp1d is not None:
            if self.x_log:
                ratio = self.x[1:]/self.x[:-1]
                if not np.allclose(ratio, ratio[0]):
                    raise GalSimIncompatibleValuesError(
                        "Cannot use a galsim.Interpolant with x_log=True unless log(x) is "
                        "equally spaced.",
                        interpolant=interpolant, x_log=x_log, x=x)
            else:
                dx = np.diff(self.x)
                if not np.allclose(dx, dx[0]):
                    raise GalSimIncompatibleValuesError(
                        "Cannot use a galsim.Interpolant with x_log=False unless x is "
                        "equally spaced.",
                        interpolant=interpolant, x_log=x_log, x=x)

    @lazy_property
    def _tab(self):
        # Store these as attributes, so don't need to worry about C++ layer persisting them.
        self._x = np.log(self.x) if self.x_log else self.x
        self._f = np.log(self.f) if self.f_log else self.f

        _x = self._x.__array_interface__['data'][0]
        _f = self._f.__array_interface__['data'][0]
        if self._interp1d is not None:
            return _galsim.LookupTable(_x, _f, len(self._x), self._interp1d._i)
        else:
            return _galsim.LookupTable(_x, _f, len(self._x), self.interpolant)

    @property
    def x_min(self):
        """The minimum x value in the lookup table.
        """
        return self._x_min

    @property
    def x_max(self):
        """The maximum x value in the lookup table.
        """
        return self._x_max

    def __len__(self): return len(self.x)

    def __call__(self, x):
        """Interpolate the `LookupTable` to get ``f(x)`` at some ``x`` value(s).

        When the `LookupTable` object is called with a single argument, it returns the value at that
        argument.  An exception will be thrown automatically if the ``x`` value is outside the
        range of the original tabulated values.  The value that is returned is the same type as
        that provided as an argument, e.g., if a single value ``x`` is provided then a single value
        of ``f`` is returned; if a tuple of ``x`` values is provided then a tuple of ``f`` values
        is returned; and so on.  Even if interpolation was done using the ``x_log`` option, the
        user should still provide ``x`` rather than ``log(x)``.

        Parameters:
            x:      The ``x`` value(s) for which ``f(x)`` should be calculated via interpolation on
                    the original ``(x,f)`` lookup table.  ``x`` can be a single float/double, or a
                    tuple, list, or arbitrarily shaped 1- or 2-dimensional NumPy array.

        Returns:
            the interpolated ``f(x)`` value(s).
        """
        orig_x = x
        # Handle the log(x) if necessary
        if self.x_log:
            x = np.log(x)

        x = np.asarray(x, dtype=float)
        try:
            if x.shape == ():
                f = self._tab.interp(float(x))
            else:
                dimen = len(x.shape)
                if dimen > 1:
                    f = np.empty_like(x.ravel(), dtype=float)
                    xx = x.astype(float,copy=False).ravel()
                    _xx = xx.__array_interface__['data'][0]
                    _f = f.__array_interface__['data'][0]
                    self._tab.interpMany(_xx, _f, len(xx))
                    f = f.reshape(x.shape)
                else:
                    f = np.empty_like(x, dtype=float)
                    xx = x.astype(float,copy=False)
                    _xx = xx.__array_interface__['data'][0]
                    _f = f.__array_interface__['data'][0]
                    self._tab.interpMany(_xx, _f, len(xx))
        except RuntimeError:
            # If there were points outside the valid range, this will have raised an exception.
            # so call _check_range to give a better error message.
            self._check_range(orig_x)
            raise  # pragma: no cover (shouldn't be able to reach here, but just in case.)

        # Handle the log(f) if necessary
        if self.f_log:
            f = np.exp(f)
        return f

    def integrate(self, x_min=None, x_max=None):
        r"""Calculate an estimate of the integral of the tabulated function from x_min to x_max:

        .. math::

            \int_{x_\mathrm{min}}^{x_\mathrm{max}} f(x) dx

        This function is not implemented for LookupTables that use log for either x or f,
        or that use a ``galsim.Interpolant``.  Also, if x_min or x_max are beyond the range
        of the tabulated function, the function will be considered to be zero there.

        .. note::

            The simplest version of this function is equivalent in functionality to the numpy
            ``trapz`` function.  However, it is usually significantly faster.  If you have a
            time-critical integration for which you are currently using ``np.trapz``::

                >>> ans = np.trapz(f, x)

            the following replacement may be faster::

                >>> ans = galsim.trapz(f, x)

            which is an alias for::

                >>> ans = galsim._LookupTable(x, f, 'linear').integrate()

        Parameters:
            x_min:      The minimum abscissa to use for the integral.  [default: None, which
                        means to use self.x_min]
            x_max:      The maximum abscissa to use for the integral.  [default: None, which
                        means to use self.x_max]

        Returns:
            an estimate of the integral
        """
        if self.x_log:
            raise GalSimNotImplementedError("log x spacing not implemented yet.")
        if self.f_log:
            raise GalSimNotImplementedError("log f values not implemented yet.")
        if not isinstance(self.interpolant, basestring):
            raise GalSimNotImplementedError(
                "Integration with interpolant=%s is not implemented."%(self.interpolant))
        if x_min is None:
            x_min = self.x_min
        else:
            x_min = max(x_min, self.x_min)
        if x_max is None:
            x_max = self.x_max
        else:
            x_max = min(x_max, self.x_max)

        if x_min < x_max:
            return self._tab.integrate(x_min, x_max)
        elif x_min == x_max:
            return 0.
        else:
            return -self.integrate(x_max, x_min)

    def integrate_product(self, g, x_min=None, x_max=None, x_factor=1.):
        r"""Calculate an estimate of the integral of the tabulated function multiplied by a second
        function from x_min to x_max:

        .. math::

            \int_{x_\mathrm{min}}^{x_\mathrm{max}} f(x) g(x) dx

        If the second function, :math:`g(x)`, is another `LookupTable`, then the quadrature will
        use the abscissae from both that function and :math:`f(x)` (i.e. ``self``).
        Otherwise, the second function will be evaluated at the abscissae of :math:`f(x)`.

        This function is not implemented for LookupTables that use log for either x or f,
        or that use a ``galsim.Interpolant``.  Also, if x_min or x_max are beyond the range
        of either tabulated function, the function will be considered to be zero there.

        Also, the second function :math:`g(x)` is always approximated with linear interpolation
        between the abscissae, even if it is a `LookupTable` with a different specified
        interpolation.

        Parameters:
            g:          The function to multiply by the current function for the integral.
            x_min:      The minimum abscissa to use for the integral.  [default: None, which
                        means to use self.x_min]
            x_max:      The maximum abscissa to use for the integral.  [default: None, which
                        means to use self.x_max]
            x_factor:   Optionally scale the x values of f by this factor when doing the integral.
                        I.e. Find :math:`\int f(x x_\mathrm{factor}) g(x) dx`. [default: 1]

        Returns:
            an estimate of the integral
        """
        from .utilities import merge_sorted
        if self.x_log:
            raise GalSimNotImplementedError("log x spacing not implemented yet.")
        if self.f_log:
            raise GalSimNotImplementedError("log f values not implemented yet.")
        if not isinstance(self.interpolant, basestring):
            raise GalSimNotImplementedError(
                "Integration with interpolant=%s is not implemented."%(self.interpolant))
        if x_min is None:
            x_min = self.x_min / x_factor
        else:
            x_min = max(x_min, self.x_min / x_factor)
        if x_max is None:
            x_max = self.x_max / x_factor
        else:
            x_max = min(x_max, self.x_max / x_factor)
        if x_min > x_max:
            return -self.integrate_product(g, x_max, x_min, x_factor)
        elif x_min == x_max:
            return 0.

        if isinstance(g, LookupTable):
            x_min = max(x_min, g.x_min)
            x_max = min(x_max, g.x_max)
            if x_min >= x_max:
                return 0.
        else:
            gx = self.x / x_factor
            gx = gx[(gx >= x_min) & (gx <= x_max)]
            gx = merge_sorted([gx, [x_min, x_max]])
            # Let this raise an appropriate error if g is not a valid function over this domain.
            gf = g(gx)
            # If g is a constant function (like lambda wave: 1), then this doesn't return
            # an array.  Make it one.
            try:
                len(gf)
            except TypeError:
                gf1 = gf
                gf = np.empty_like(gx, dtype=float)
                gf.fill(gf1)
            g = _LookupTable(gx, gf, 'linear')

        return self._tab.integrate_product(g._tab, float(x_min), float(x_max), float(x_factor))

    def _check_range(self, x):
        slop = (self.x_max - self.x_min) * 1.e-6
        if np.min(x,initial=self.x_min) < self.x_min - slop:
            xx = np.array(x)
            raise GalSimRangeError("x value(s) below the range of the LookupTable.",
                                   xx[xx<self.x_min], self.x_min, self.x_max) from None
        if np.max(x,initial=self.x_max) > self.x_max + slop:  # pragma: no branch
            xx = np.array(x)
            raise GalSimRangeError("x value(s) above the range of the LookupTable.",
                                   xx[xx>self.x_max], self.x_min, self.x_max) from None

    def getArgs(self):
        return self.x

    def getVals(self):
        return self.f

    def getInterp(self):
        return self.interpolant

    def isLogX(self):
        return self.x_log

    def isLogF(self):
        return self.f_log

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, LookupTable) and
                 np.array_equal(self.x,other.x) and
                 np.array_equal(self.f,other.f) and
                 self.x_log == other.x_log and
                 self.f_log == other.f_log and
                 self.interpolant == other.interpolant))
    def __ne__(self, other): return not self.__eq__(other)

    def __hash__(self):
        # Cache this in case self.x, self.f are long.
        if not hasattr(self, '_hash'):
            self._hash = hash(("galsim.LookupTable", tuple(self.x), tuple(self.f), self.x_log,
                               self.f_log, self.interpolant))
        return self._hash


    def __repr__(self):
        return 'galsim.LookupTable(x=array(%r), f=array(%r), interpolant=%r, x_log=%r, f_log=%r)'%(
            self.x.tolist(), self.f.tolist(), self.interpolant, self.x_log, self.f_log)

    def __str__(self):
        s = 'galsim.LookupTable(x=%s, f=%s'%(_str_array(self.x), _str_array(self.f))
        if self.interpolant != 'spline':
            s += ', interpolant=%r'%(self.interpolant)
        if self.x_log:
            s += ', x_log=True'
        if self.f_log:
            s += ', f_log=True'
        s += ')'
        return s

    @classmethod
    def from_file(cls, file_name, interpolant='spline', x_log=False, f_log=False, amplitude=1.0):
        """Create a `LookupTable` from a file of x, f values.

        This reads in a file, which should contain two columns with the x and f values.

        Parameters:
            file_name:      A file from which to read the ``(x,f)`` pairs.
            interpolant:    Type of interpolation to use. [default: 'spline']
            x_log:          Whether the x values should be uniform in log rather than lienar.
                            [default: False]
            f_log:          Whether the f values should be interpolated using their logarithms
                            rather than their raw values. [default: False]
            amplitude:      An optional scaling of the f values relative to the values in the file
                            [default: 1.0]
        """
        # We don't require pandas as a dependency, but if it's available, this is much faster.
        # cf. http://stackoverflow.com/questions/15096269/the-fastest-way-to-read-input-in-python
        ParserError = AttributeError # In case we don't get to the line below where we import
                                     # it from pandas.
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                import pandas
                from pandas.errors import ParserError
            data = pandas.read_csv(file_name, comment='#', sep=r'\s+', header=None)
            data = data.values.transpose()
        except (ImportError, AttributeError, ParserError):
            data = np.loadtxt(file_name).transpose()
        if data.shape[0] != 2:
            raise GalSimValueError("File provided for LookupTable does not have 2 columns",
                                   file_name)
        x=data[0]
        f=data[1]
        if amplitude != 1.0:
            f[:] *= amplitude
        return LookupTable(x, f, interpolant=interpolant, x_log=x_log, f_log=f_log)

    @classmethod
    def from_func(cls, func, x_min, x_max, npoints=2000, interpolant='spline',
                  x_log=False, f_log=False):
        """Create a `LookupTable` from a callable function

        This constructs a `LookupTable` over the given range from x_min and x_max, calculating the
        corresponding f values from the given function (technically any callable object).

        Parameters:
            func:           A callable function.
            x_min:          The minimum x value at which to evalue the function and store in the
                            lookup table.
            x_max:          The maximum x value at which to evalue the function and store in the
                            lookup table.
            npoints:        Number of x values at which to evaluate the function. [default: 2000]
            interpolant:    Type of interpolation to use. [default: 'spline']
            x_log:          Whether the x values should be uniform in log rather than lienar.
                            [default: False]
            f_log:          Whether the f values should be interpolated using their logarithms
                            rather than their raw values. [default: False]
        """
        if x_log:
            x = np.exp(np.linspace(np.log(x_min), np.log(x_max), npoints))
        else:
            x = np.linspace(x_min, x_max, npoints)
        f = np.array([func(xx) for xx in x], dtype=float)
        return cls(x, f, interpolant=interpolant, x_log=x_log, f_log=f_log)

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('_tab',None)
        return d

    def __setstate__(self, d):
        self.__dict__ = d

def _LookupTable(x, f, interpolant='spline', x_log=False, f_log=False):
    """Make a `LookupTable` but without using any of the sanity checks or array manipulation used
    in the normal initializer.

    The input x values must be already sorted.

    Parameters:
        x:              Strictly increasing NumPy array of ``x`` values.
        f:              NumPy array of ``f(x)`` values.
        interpolant:    Type of interpolation to use, with the options being 'floor', 'ceil',
                        'nearest', 'linear', 'spline', or a `galsim.Interpolant` or string
                        convertible to one.  [default: 'spline']
        x_log:          Set to True if you wish to interpolate using log(x) rather than x.  Note
                        that all inputs / outputs will still be x, it's just a question of how the
                        interpolation is done. [default: False]
        f_log:          Set to True if you wish to interpolate using log(f) rather than f.  Note
                        that all inputs / outputs will still be f, it's just a question of how the
                        interpolation is done. [default: False]
    """
    ret = LookupTable.__new__(LookupTable)
    ret.x = np.ascontiguousarray(x, dtype=float)
    ret.f = np.ascontiguousarray(f, dtype=float)
    ret.interpolant = interpolant
    ret.x_log = x_log
    ret.f_log = f_log
    ret._x_min = ret.x[0]
    ret._x_max = ret.x[-1]
    if interpolant in ('nearest', 'linear', 'ceil', 'floor', 'spline'):
        ret._interp1d = None
    else:
        ret._interp1d = convert_interpolant(interpolant)
    return ret


def trapz(f, x, interpolant='linear'):
    """Integrate f(x) using the trapezoidal rule.

    Equivalent to np.trapz(f,x) for 1d array inputs.  Intended as a drop-in replacement,
    which is usually faster.

    Parameters:
        f:              The ordinates of the function to integrate.
        x:              The abscissae of the function to integrate.
        interpolant:    The interpolant to use between the tabulated points.  [default: 'linear',
                        which matches the behavior of np.trapz]

    Returns:
        Estimate of the integral.
    """
    if len(x) >= 2:
        return _LookupTable(x, f, interpolant=interpolant).integrate()
    else:
        return 0.


class LookupTable2D:
    """
    LookupTable2D represents a 2-dimensional lookup table to store function values that may be slow
    to calculate, for which interpolating from a lookup table is sufficiently accurate.  A
    LookupTable2D is also useful for evaluating periodic 2-d functions given samples from a single
    period.

    A LookupTable2D representing the function f(x, y) may be constructed from a list or array of
    ``x`` values, a list or array of ``y`` values, and a 2D array of function evaluations at all
    combinations of x and y values.  For instance::

        >>> x = np.arange(5)
        >>> y = np.arange(8)
        >>> z = x + y[:, np.newaxis]  # function is x + y, dimensions of z are (8, 5)
        >>> tab2d = galsim.LookupTable2D(x, y, z)

    To evaluate new function values with the lookup table, use the () operator::

        >>> print tab2d(2.2, 3.3)
        5.5

    The () operator can also accept sequences (lists, tuples, numpy arrays, ...) for the x and y
    arguments at which to evaluate the LookupTable2D.  Normally, the x and y sequences should have
    the same length, which will also be the length of the output sequence::

        >>> print tab2d([1, 2], [3, 5])
        array([ 4., 7.])

    If you add ``grid=True`` as an additional kwarg, however, then the () operator will generate
    interpolated values at the outer product of x-values and y-values.  So in this case, the x and
    y sequences can have different lengths Nx and Ny, and the result will be a 2D array with
    dimensions (Nx, Ny)::

        >>> print tab2d([1, 2], [3, 5], grid=True)
        array([[ 4., 6.],
               [ 5., 7.]])

    The default interpolation method is linear.  Other choices for the interpolant are:

        - 'floor'
        - 'ceil'
        - 'nearest'
        - 'spline' (a Catmull-Rom cubic spline).
        - a `galsim.Interpolant` or string convertible to one.

    ::

        >>> tab2d = galsim.LookupTable2D(x, y, z, interpolant='floor')
        >>> tab2d(2.2, 3.7)
        5.0
        >>> tab2d = galsim.LookupTable2D(x, y, z, interpolant='ceil')
        >>> tab2d(2.2, 3.7)
        7.0
        >>> tab2d = galsim.LookupTable2D(x, y, z, interpolant='nearest')
        >>> tab2d(2.2, 3.7)
        6.0

    For interpolant='spline' or a `galsim.Interpolant`, the input arrays must be uniformly spaced.
    For interpolant='spline', the derivatives df / dx, df / dy, and d^2 f / dx dy at grid-points may
    also optionally be provided if they're known, which will generally yield a more accurate
    interpolation (these derivatives will be estimated from finite differences if they're not
    provided).

    The ``edge_mode`` keyword describes how to handle extrapolation beyond the initial input range.
    Possibilities include:

      - 'raise': raise an exception.  (This is the default.)
      - 'warn': issues a warning, then falls back to edge_mode='constant'.
      - 'constant': Return a constant specified by the ``constant`` keyword.
      - 'wrap': infinitely wrap the initial range in both directions.

    In order for LookupTable2D to determine the wrapping period when edge_mode='wrap', either the
    x and y grid points need to be equally spaced (in which case the x-period is inferred as
    len(x)*(x[1]-x[0]) and similarly for y), or the first/last row/column of f must be identical,
    in which case the x-period is inferred as x[-1] - x[0].  (If both conditions are satisfied
    (equally-spaced x and y and identical first/last row/column of f, then the x-period is inferred
    as len(x)*(x[1]-x[0]))::

        >>> x = np.arange(5)
        >>> y = np.arange(8)
        >>> z = x + y[:, np.newaxis]  # function is x + y, dimensions of z are (8, 5)
        >>> tab2d = galsim.LookupTable2D(x, y, z, edge_mode='raise')
        >>> tab2d(7, 7)
        ValueError: Extrapolating beyond input range.

        >>> tab2d = galsim.LookupTable2D(x, y, z, edge_mode='constant', constant=1.0)
        1.0

        >>> tab2d = galsim.LookupTable2D(x, y, z, edge_mode='wrap')
        ValueError: Cannot wrap `f` array with unequal first/last column/row.

    We extend the x and y arrays with a uniform spacing, though any monotonic spacing would work.
    Note that the [(0,1), (0,1)] argument in np.pad below extends the z array by 0 rows/columns in
    the leading direction, and 1 row/column in the trailing direction::

        >>> x = np.append(x, x[-1] + (x[-1]-x[-2]))
        >>> y = np.append(y, y[-1] + (y[-1]-y[-2]))
        >>> z = np.pad(z, [(0,1), (0,1)], mode='wrap')
        >>> tab2d = galsim.LookupTable2D(x, y, z, edge_mode='wrap')
        >>> tab2d(2., 2.)
        4.0
        >>> tab2d(2.+5, 2.)  # The period is 5 in the x direction
        4.0
        >>> tab2d(2.+3*5, 2.+4*8)  # The period is 8 in the y direction
        4.0

    Parameters:
        x:              Strictly increasing array of ``x`` positions at which to create table.
        y:              Strictly increasing array of ``y`` positions at which to create table.
        f:              Nx by Ny input array of function values.
        dfdx:           Optional first derivative of f wrt x.  Only used if interpolant='spline'.
                        [default: None]
        dfdy:           Optional first derivative of f wrt y.  Only used if interpolant='spline'.
                        [default: None]
        d2fdxdy:        Optional cross derivative of f wrt x and y.  Only used if
                        interpolant='spline'.  [default: None]
        interpolant:    Type of interpolation to use.  One of 'floor', 'ceil', 'nearest', 'linear',
                        'spline', or a `galsim.Interpolant` or string convertible to one.
                        [default: 'linear']
        edge_mode:      Keyword controlling how extrapolation beyond the input range is handled.
                        See above for details.  [default: 'raise']
        constant:       A constant to return when extrapolating beyond the input range and
                        ``edge_mode='constant'``.  [default: 0]
    """
    def __init__(self, x, y, f, dfdx=None, dfdy=None, d2fdxdy=None,
                 interpolant='linear', edge_mode='raise', constant=0):
        if edge_mode not in ('raise', 'warn', 'wrap', 'constant'):
            raise GalSimValueError("Unknown edge_mode.", edge_mode,
                                   ('raise', 'warn', 'wrap', 'constant'))

        x = np.asarray(x, dtype=float)
        y = np.asarray(y, dtype=float)
        f = np.asarray(f, dtype=float)

        dx = np.diff(x)
        dy = np.diff(y)
        equal_spaced = np.allclose(dx, dx[0]) and np.allclose(dy, dy[0])

        if not all(dx > 0):
            raise GalSimValueError("x input grids is not strictly increasing.", x)
        if not all(dy > 0):
            raise GalSimValueError("y input grids is not strictly increasing.", y)

        fshape = f.shape
        if fshape != (len(y), len(x)):
            raise GalSimIncompatibleValuesError(
                "Shape of f incompatible with lengths of x,y", f=f, x=x, y=y)

        # Check if interpolant is a string that we understand.  If not, try convert_interpolant
        if interpolant in ('nearest', 'linear', 'ceil', 'floor', 'spline'):
            self._interp2d = None
            padrange = 2 if interpolant == 'spline' else 1
        else:
            self._interp2d = convert_interpolant(interpolant)
            padrange = int(np.ceil(self._interp2d.xrange))
        self.interpolant = interpolant

        # Check if need equal-spaced arrays
        if (self._interp2d is not None or interpolant  == 'spline'):
            if not equal_spaced:
                raise GalSimIncompatibleValuesError(
                "Cannot use a galsim.Interpolant in LookupTable2D unless x and y are "
                "equally spaced.", interpolant=interpolant, x=x, y=y)

        self.edge_mode = edge_mode
        self.constant = float(constant)

        if self.edge_mode == 'wrap':
            # Can wrap if x and y arrays are equally spaced ...
            if equal_spaced:
                if 2*padrange > len(x) or 2*padrange > len(y):
                    raise GalSimValueError(
                        "Cannot wrap an image which is smaller than the Interpolant",
                        (x, y, interpolant))
                # Underlying Table2D requires us to extend x, y, and f.
                self.xperiod = x[-1]-x[0]+dx[0]
                self.yperiod = y[-1]-y[0]+dy[0]
                self.x0 = x[0]
                self.y0 = y[0]
                x = np.hstack([x[0]-np.cumsum([dx[0]]*(padrange-1)),
                               x,
                               x[-1]+np.cumsum([dx[0]]*padrange)])
                y = np.hstack([y[0]-np.cumsum([dy[0]]*(padrange-1)),
                               y,
                               y[-1]+np.cumsum([dy[0]]*padrange)])
                f = np.pad(f, [(padrange-1, padrange)]*2, mode='wrap')
            # Can also wrap non-uniform grids if edges match
            elif (all(f[0] == f[-1]) and all(f[:,0] == f[:,-1])):
                self.x0 = x[0]
                self.y0 = y[0]
                self.xperiod = x[-1] - x[0]
                self.yperiod = y[-1] - y[0]
            else:
                raise GalSimIncompatibleValuesError(
                    "Cannot use edge_mode='wrap' unless either x and y are equally "
                    "spaced or first/last row/column of f are identical.",
                    edge_mode=edge_mode, x=x, y=y, f=f)

        self.x = np.ascontiguousarray(x)
        self.y = np.ascontiguousarray(y)
        self.f = np.ascontiguousarray(f)

        der_exist = [kw is not None for kw in [dfdx, dfdy, d2fdxdy]]
        if self.interpolant == 'spline':
            if any(der_exist):
                if not all(der_exist):
                    raise GalSimIncompatibleValuesError(
                        "Must specify all of dfdx, dfdy, d2fdxdy if one is specified",
                        dfdx=dfdx, dfdy=dfdy, d2fdxdy=d2fdxdy)
            else:
                # Use finite differences if derivatives not provided
                dfdx = np.empty_like(f)
                diffx = self.x[2:] - self.x[:-2]
                dfdx[:, 1:-1] = (f[:, 2:] - f[:, :-2])/diffx
                dfdx[:, 0] = (f[:, 1] - f[:, 0])/dx[0]
                dfdx[:, -1] = (f[:, -1] - f[:, -2])/dx[-1]

                dfdy = np.empty_like(f)
                diffy = self.y[2:] - self.y[:-2]
                dfdy[1:-1, :] = (f[2:, :] - f[:-2, :])/diffy[:,None]
                dfdy[0, :] = (f[1, :] - f[0, :])/dy[0]
                dfdy[-1, :] = (f[-1, :] - f[-2, :])/dy[-1]

                d2fdxdy = np.empty_like(f)
                d2fdxdy[1:-1, :] = (dfdx[2:, :] - dfdx[:-2, :])/diffy[:,None]
                d2fdxdy[0, :] = (dfdx[1, :] - dfdx[0, :])/dy[0]
                d2fdxdy[-1, :] = (dfdx[-1, :] - dfdx[-2, :])/dy[-1]
        else:
            if any(der_exist):
                raise GalSimIncompatibleValuesError(
                    "Only specify dfdx, dfdy, d2fdxdy if interpolant is 'spline'.",
                    dfdx=dfdx, dfdy=dfdy, d2fdxdy=d2fdxdy, interpolant=interpolant)

        if dfdx is not None:
            dfdx = np.ascontiguousarray(dfdx, dtype=float)
            dfdy = np.ascontiguousarray(dfdy, dtype=float)
            d2fdxdy = np.ascontiguousarray(d2fdxdy, dtype=float)

            if dfdx.shape != f.shape or dfdy.shape != f.shape or d2fdxdy.shape != f.shape:
                raise GalSimIncompatibleValuesError(
                    "derivative shapes must match f shape",
                    dfdx=dfdx, dfdy=dfdy, d2fdxdy=d2fdxdy)

        self.dfdx = dfdx
        self.dfdy = dfdy
        self.d2fdxdy = d2fdxdy

    @lazy_property
    def _tab(self):
        _x = self.x.__array_interface__['data'][0]
        _y = self.y.__array_interface__['data'][0]
        _f = self.f.__array_interface__['data'][0]
        if self._interp2d is not None:
            return _galsim.LookupTable2D(_x, _y, _f, len(self.x), len(self.y),
                                         self._interp2d._i)
        elif self.interpolant == 'spline':
            _dfdx = self.dfdx.__array_interface__['data'][0]
            _dfdy = self.dfdy.__array_interface__['data'][0]
            _d2fdxdy = self.d2fdxdy.__array_interface__['data'][0]
            return _galsim.LookupTable2D(_x, _y, _f, len(self.x), len(self.y),
                                         _dfdx, _dfdy, _d2fdxdy)
        else:
            return _galsim.LookupTable2D(_x, _y, _f, len(self.x), len(self.y),
                                         self.interpolant)

    def getXArgs(self):
        return self.x

    def getYArgs(self):
        return self.y

    def getVals(self):
        return self.f

    def _inbounds(self, x, y):
        """Return whether or not *all* coords specified by x and y are in bounds of the original
        interpolated array."""
        # Only used if edge_mode != 'wrap', so original x/y arrays are unmodified.
        return (np.min(x) >= self.x[0] and np.max(x) <= self.x[-1] and
                np.min(y) >= self.y[0] and np.max(y) <= self.y[-1])

    def _wrap_args(self, x, y):
        """Wrap points back into the fundamental period."""
        # Original x and y may have been modified, so need to use x0 and xperiod attributes here.
        #x = (x-self.x0) % self.xperiod + self.x0
        #y = (y-self.y0) % self.yperiod + self.y0
        _x = x.__array_interface__['data'][0]
        _y = y.__array_interface__['data'][0]
        _galsim.WrapArrayToPeriod(_x, len(x), self.x0, self.xperiod)
        _galsim.WrapArrayToPeriod(_y, len(y), self.y0, self.yperiod)
        return x, y

    @property
    def _bounds(self):
        # Only meaningful if edge_mode is 'raise' or 'warn', in which case original x/y arrays are
        # unmodified.
        return BoundsD(self.x[0], self.x[-1], self.y[0], self.y[-1])

    def _call_inbounds(self, x, y, grid=False):
        _x = x.__array_interface__['data'][0]
        _y = y.__array_interface__['data'][0]
        if grid:
            f = np.empty((len(y), len(x)), dtype=float)
            _f = f.__array_interface__['data'][0]
            self._tab.interpGrid(_x, _y, _f, len(x), len(y))
            return f
        else:
            f = np.empty_like(x, dtype=float)
            _f = f.__array_interface__['data'][0]
            self._tab.interpMany(_x, _y, _f, len(x))
            return f

    def _call_constant(self, x, y, grid=False):
        x = np.asarray(x, dtype=float)
        y = np.asarray(y, dtype=float)
        if grid:
            f = np.empty((len(y), len(x)), dtype=float)
            # Fill in interpolated values first, then go back and fill in
            # constants
            _x = x.__array_interface__['data'][0]
            _y = y.__array_interface__['data'][0]
            _f = f.__array_interface__['data'][0]
            self._tab.interpGrid(_x, _y, _f, len(x), len(y))
            badx = (x < self.x[0]) | (x > self.x[-1])
            bady = (y < self.y[0]) | (y > self.y[-1])
            f[bady, :] = self.constant
            f[:, badx] = self.constant
            return f
        else:
            # Start with constant array, then interpolate good positions
            f = np.empty_like(x, dtype=float)
            f.fill(self.constant)
            good = ((x >= self.x[0]) & (x <= self.x[-1]) &
                    (y >= self.y[0]) & (y <= self.y[-1]))
            xx = np.ascontiguousarray(x[good].ravel(), dtype=float)
            yy = np.ascontiguousarray(y[good].ravel(), dtype=float)
            tmp = np.empty_like(xx, dtype=float)
            _xx = xx.__array_interface__['data'][0]
            _yy = yy.__array_interface__['data'][0]
            _tmp = tmp.__array_interface__['data'][0]
            self._tab.interpMany(_xx, _yy, _tmp, len(xx))
            f[good] = tmp
            return f

    def _call_raise(self, x, y, grid=False):
        if not self._inbounds(x, y):
            pos = find_out_of_bounds_position(x, y, self._bounds, grid)
            raise GalSimBoundsError("Extrapolating beyond input range.",
                                    pos, self._bounds)
        return self._call_inbounds(x, y, grid)

    def _call_warn(self, x, y, grid=False):
        if not self._inbounds(x, y):
            pos = find_out_of_bounds_position(x, y, self._bounds, grid)
            galsim_warn("Extrapolating beyond input range. {!r} not in {!r}".format(
                        pos, self._bounds))
        return self._call_constant(x, y, grid)

    def _call_wrap(self, x, y, grid=False):
        x, y = self._wrap_args(x, y)
        return self._call_inbounds(x, y, grid)

    def __call__(self, x, y, grid=False):
        """Interpolate at an arbitrary point or points.

        Parameters:
            x:      Either a single x value or an array of x values at which to interpolate.
            y:      Either a single y value or an array of y values at which to interpolate.
            grid:   Optional boolean indicating that output should be a 2D array corresponding
                    to the outer product of input values.  If False (default), then the output
                    array will be congruent to x and y.

        Returns:
            a scalar value if x and y are scalar, or a numpy array if x and y are arrays.
        """
        if np.__version__ >= "2.0":  # pragma: no branch
            # I'm not sure if there is a simpler way to do this in 2.0.
            # We want a copy when edge_mode == wrap. Otherwise, only copy if dtype changes or
            # x,y aren't already arrays. That used to be simple...
            copy = True if self.edge_mode=='wrap' else None
        else:
            copy = self.edge_mode=='wrap'
        x1 = np.array(x, dtype=float, copy=copy)
        y1 = np.array(y, dtype=float, copy=copy)
        x2 = np.ascontiguousarray(x1.ravel(), dtype=float)
        y2 = np.ascontiguousarray(y1.ravel(), dtype=float)

        if self.edge_mode == 'raise':
            f = self._call_raise(x2, y2, grid)
        elif self.edge_mode == 'warn':
            f = self._call_warn(x2, y2, grid)
        elif self.edge_mode == 'wrap':
            f = self._call_wrap(x2, y2, grid)
        else: # constant
            f = self._call_constant(x2, y2, grid)

        if isinstance(x, numbers.Real):
            return f[0]
        else:
            if not grid:
                f = f.reshape(x1.shape)
            return f

    def _gradient_inbounds(self, x, y, grid=False):
        if grid:
            dfdx = np.empty((len(y), len(x)), dtype=float)
            dfdy = np.empty((len(y), len(x)), dtype=float)
            _x = x.__array_interface__['data'][0]
            _y = y.__array_interface__['data'][0]
            _dfdx = dfdx.__array_interface__['data'][0]
            _dfdy = dfdy.__array_interface__['data'][0]
            self._tab.gradientGrid(_x, _y, _dfdx, _dfdy, len(x), len(y))
            return dfdx, dfdy
        else:
            dfdx = np.empty_like(x)
            dfdy = np.empty_like(x)
            _x = x.__array_interface__['data'][0]
            _y = y.__array_interface__['data'][0]
            _dfdx = dfdx.__array_interface__['data'][0]
            _dfdy = dfdy.__array_interface__['data'][0]
            self._tab.gradientMany(_x, _y, _dfdx, _dfdy, len(x))
            return dfdx, dfdy

    def _gradient_raise(self, x, y, grid=False):
        if not self._inbounds(x, y):
            pos = find_out_of_bounds_position(x, y, self._bounds, grid)
            raise GalSimBoundsError("Extrapolating beyond input range.",
                                    pos, self._bounds)
        return self._gradient_inbounds(x, y, grid)

    def _gradient_warn(self, x, y, grid=False):
        if not self._inbounds(x, y):
            pos = find_out_of_bounds_position(x, y, self._bounds, grid)
            galsim_warn("Extrapolating beyond input range. {!r} not in {!r}".format(
                        pos, self._bounds))
        return self._gradient_constant(x, y, grid)

    def _gradient_wrap(self, x, y, grid=False):
        x, y = self._wrap_args(x, y)
        return self._gradient_inbounds(x, y, grid)

    def _gradient_constant(self, x, y, grid=False):
        x = np.asarray(x, dtype=float)
        y = np.asarray(y, dtype=float)
        if grid:
            dfdx = np.empty((len(y), len(x)), dtype=float)
            dfdy = np.empty((len(y), len(x)), dtype=float)
            _x = x.__array_interface__['data'][0]
            _y = y.__array_interface__['data'][0]
            _dfdx = dfdx.__array_interface__['data'][0]
            _dfdy = dfdy.__array_interface__['data'][0]
            self._tab.gradientGrid(_x, _y, _dfdx, _dfdy, len(x), len(y))
            badx = (x < self.x[0]) | (x > self.x[-1])
            bady = (y < self.y[0]) | (y > self.y[-1])
            dfdx[bady,:] = 0.0
            dfdx[:, badx] = 0.0
            dfdy[bady,:] = 0.0
            dfdy[:, badx] = 0.0
            return dfdx, dfdy
        else:
            dfdx = np.empty_like(x, dtype=float)
            dfdy = np.empty_like(x, dtype=float)
            dfdx.fill(0.0)
            dfdy.fill(0.0)
            good = ((x >= self.x[0]) & (x <= self.x[-1]) &
                    (y >= self.y[0]) & (y <= self.y[-1]))
            x = np.ascontiguousarray(x[good].ravel(), dtype=float)
            y = np.ascontiguousarray(y[good].ravel(), dtype=float)
            tmp1 = np.empty_like(x, dtype=float)
            tmp2 = np.empty_like(x, dtype=float)
            _x = x.__array_interface__['data'][0]
            _y = y.__array_interface__['data'][0]
            _tmp1 = tmp1.__array_interface__['data'][0]
            _tmp2 = tmp2.__array_interface__['data'][0]
            self._tab.gradientMany(_x, _y, _tmp1, _tmp2, len(x))
            dfdx[good] = tmp1
            dfdy[good] = tmp2
            return dfdx, dfdy

    def gradient(self, x, y, grid=False):
        """Calculate the gradient of the function at an arbitrary point or points.

        Parameters:
            x:      Either a single x value or an array of x values at which to compute
                    the gradient.
            y:      Either a single y value or an array of y values at which to compute
                    the gradient.
            grid:   Optional boolean indicating that output should be a 2-tuple of 2D arrays
                    corresponding to the outer product of input values.  If False (default),
                    then the output arrays will be congruent to x and y.

        Returns:
            A tuple of (dfdx, dfdy) where dfdx, dfdy are single values (if x,y were single
            values) or numpy arrays.
        """
        if np.__version__ >= "2.0":  # pragma: no branch
            copy = True if self.edge_mode=='wrap' else None
        else:
            copy = self.edge_mode=='wrap'
        x1 = np.array(x, dtype=float, copy=copy)
        y1 = np.array(y, dtype=float, copy=copy)
        x2 = np.ascontiguousarray(x1.ravel(), dtype=float)
        y2 = np.ascontiguousarray(y1.ravel(), dtype=float)

        if self.edge_mode == 'raise':
            dfdx, dfdy = self._gradient_raise(x2, y2, grid)
        if self.edge_mode == 'warn':
            dfdx, dfdy = self._gradient_warn(x2, y2, grid)
        elif self.edge_mode == 'wrap':
            dfdx, dfdy = self._gradient_wrap(x2, y2, grid)
        else: # constant
            dfdx, dfdy = self._gradient_constant(x2, y2, grid)

        if isinstance(x, numbers.Real):
            return dfdx[0], dfdy[0]
        else:
            if not grid:
                dfdx = dfdx.reshape(x1.shape)
                dfdy = dfdy.reshape(x1.shape)
            return dfdx, dfdy

    def __str__(self):
        return "galsim.LookupTable2D(x=%s, y=%s, f=[%s,...,%s], interpolant=%r, edge_mode=%r)"%(
            _str_array(self.x), _str_array(self.y),
            _str_array(self.f[0]), _str_array(self.f[-1]),
            self.interpolant, self.edge_mode)

    def __repr__(self):
        return ("galsim.LookupTable2D(x=array(%r), y=array(%r), "
                "f=array(%r), interpolant=%r, edge_mode=%r, constant=%r)"%(
            self.x.tolist(), self.y.tolist(), self.f.tolist(), self.interpolant, self.edge_mode,
            self.constant))

    def __eq__(self, other):
        if self is other: return True
        if not (isinstance(other, LookupTable2D) and
                np.array_equal(self.x,other.x) and
                np.array_equal(self.y,other.y) and
                np.array_equal(self.f,other.f) and
                self.interpolant == other.interpolant and
                self.edge_mode == other.edge_mode and
                self.constant == other.constant):
            return False
        else:
            if self.interpolant == 'spline':
                return (np.array_equal(self.dfdx, other.dfdx) and
                        np.array_equal(self.dfdy, other.dfdy) and
                        np.array_equal(self.d2fdxdy, other.d2fdxdy))
            return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        if not hasattr(self, '_hash'):
            self._hash = hash(("galsim.LookupTable2D", tuple(self.x.ravel()), tuple(self.y.ravel()),
                               tuple(self.f.ravel()), self.interpolant, self.edge_mode,
                               self.constant,
                               tuple(self.dfdx.ravel()) if self.dfdx is not None else None,
                               tuple(self.dfdy.ravel()) if self.dfdy is not None else None,
                               tuple(self.d2fdxdy.ravel()) if self.d2fdxdy is not None else None))
        return self._hash

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('_tab',None)
        return d

    def __setstate__(self, d):
        self.__dict__ = d


def _LookupTable2D(x, y, f, interpolant, edge_mode, constant,
                   dfdx=None, dfdy=None, d2fdxdy=None,
                   x0=None, y0=None, xperiod=None, yperiod=None):
    """Make a `LookupTable2D` but without using any of the sanity checks or array manipulation used
    in the normal initializer.
    """
    ret = LookupTable2D.__new__(LookupTable2D)
    ret.x = x
    ret.y = y
    ret.f = f
    ret.interpolant = interpolant
    ret.edge_mode = edge_mode
    ret.constant = constant
    ret.dfdx = dfdx
    ret.dfdy = dfdy
    ret.d2fdxdy = d2fdxdy
    ret.x0 = x0
    ret.y0 = y0
    ret.xperiod = xperiod
    ret.yperiod = yperiod
    if interpolant in ('nearest', 'linear', 'ceil', 'floor', 'spline'):
        ret._interp2d = None
    else:
        ret._interp2d = convert_interpolant(interpolant)
    return ret

def find_out_of_bounds_position(x, y, bounds, grid=False):
    """Given arrays of x and y values that are known to contain at least one
    position that is out-of-bounds of the given bounds instance, return one
    such PositionD.

    Parameters:
        x:          Array of x values
        y:          Array of y values
        bounds:     `Bounds` instance
        grid:       Bool indicating whether to check the outer product of x and y
                    (grid=True), or each sequential pair of x and y (grid=False).
                    If the latter, then x and y should have the same shape.

    Returns:
        a `PositionD` from x and y that is out-of-bounds of bounds.
    """
    if grid:
        # It's enough to check corners for grid input
        for x_ in (np.min(x), np.max(x)):
            for y_ in (np.min(y), np.max(y)):
                if (x_ < bounds.xmin or x_ > bounds.xmax or
                    y_ < bounds.ymin or y_ > bounds.ymax):
                    return _PositionD(x_, y_)
    else:
        # Faster to check all points than to iterate through them one-by-one?
        w = np.where((x < bounds.xmin) | (x > bounds.xmax) |
                     (y < bounds.ymin) | (y > bounds.ymax))
        if len(w[0]) > 0:
            return _PositionD(x[w[0][0]], y[w[0][0]])
    raise GalSimError("No out-of-bounds position")




# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

from .extra import ExtraOutputBuilder, RegisterExtraOutput
from ..image import ImageS

# The badpix extra output type is currently just a placeholder for when we eventually add
# defects, saturation, etc.  Now it always just builds an Image with all 0's.

class BadPixBuilder(ExtraOutputBuilder):
    """This builds a bad pixel mask image to go along with each regular data image.

    There's not much here currently, since GalSim doesn't yet have any image artifacts that
    would be appropriate to do something with here.  So this is mostly just a placeholder for
    when we eventually add defects, saturation, etc.
    """

    # The function to call at the end of building each stamp
    def processStamp(self, obj_num, config, base, logger):
        # Note: This is just a placeholder for now.  Once we implement defects, saturation, etc.,
        # these features should be marked in the badpix mask.  For now though, all pixels = 0.
        if base['do_noise_in_stamps']:
            badpix_im = ImageS(base['current_stamp'].bounds, wcs=base['wcs'], init_value=0)
            self.scratch[obj_num] = badpix_im

    # The function to call at the end of building each image
    def processImage(self, index, obj_nums, config, base, logger):
        image = ImageS(base['image_bounds'], wcs=base['wcs'], init_value=0)
        if len(self.scratch) > 0.:
            # If we have been accumulating the variance on the stamps, build the total from them.
            # Make sure to only use the stamps for objects in this image.
            for obj_num in obj_nums:
                stamp = self.scratch[obj_num]
                b = stamp.bounds & image.bounds
                if b.isDefined():  # pragma: no branch
                    # This next line is equivalent to:
                    #    image[b] |= stamp[b]
                    # except that this doesn't work through the proxy.  We can only call methods
                    # that don't start with _.  Hence using the more verbose form here.
                    image.setSubImage(b, image.subImage(b) | stamp[b])
        else:
            # Otherwise, build the bad pixel mask here.
            # Again, nothing here yet.
            pass
        self.data[index] = image


# Register this as a valid extra output
RegisterExtraOutput('badpix', BadPixBuilder())



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np
import os

import galsim
from galsim_test_helpers import *

path, filename = os.path.split(__file__)
imgdir = os.path.join(path, "SBProfile_comparison_images") # Directory containing the reference
                                                           # images.

@timer
def test_spergel():
    """Test the generation of a specific Spergel profile against a known result.
    """
    test_spergel_nu = [-0.85, -0.5, 0.0, 0.85, 4.0]
    mathica_enclosed_fluxes = [3.06256e-2, 9.99995e-6, 6.06443e-10, 2.94117e-11, 6.25011e-12]
    mathica_enclosing_radii = [2.3973e-17, 1.00001e-5, 1.69047e-3, 5.83138e-3, 1.26492e-2]

    for nu, enclosed_flux, enclosing_radius in zip(test_spergel_nu,
                                                   mathica_enclosed_fluxes,
                                                   mathica_enclosing_radii):
        filename = "spergel_nu{0:.2f}.fits".format(nu)
        savedImg = galsim.fits.read(os.path.join(imgdir, filename))
        savedImg.setCenter(0,0)
        dx = 0.2
        myImg = galsim.ImageF(savedImg.bounds, scale=dx)
        myImg.setCenter(0,0)

        spergel = galsim.Spergel(nu=nu, half_light_radius=1.0)
        # Reference images were made with old centering,
        # which is equivalent to use_true_center=False.
        myImg = spergel.drawImage(myImg, scale=dx, method="sb", use_true_center=False)

        np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject Spergel disagrees with expected result")

        gsp = galsim.GSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)
        spergel2 = galsim.Spergel(nu=nu, half_light_radius=1.0, gsparams=gsp)
        assert spergel2 != spergel
        assert spergel2 == spergel.withGSParams(gsp)
        assert spergel2 == spergel.withGSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)

        # nu < 0 has inf for xValue(0,0), so the x tests fail for them.
        check_basic(spergel, "Spergel with nu=%f"%nu, do_x=(nu > 0))

        # Only nu >= -0.3 give reasonably sized FFTs,
        # and small nu method='phot' is super slow.
        if nu >= -0.3:
            test_im = galsim.Image(16,16,scale=dx)
            do_kvalue(spergel,test_im, "Spergel(nu={0:1}) ".format(nu))

            # Test photon shooting.
            # Convolve with a small gaussian to smooth out the central peak.
            spergel2 = galsim.Convolve(spergel, galsim.Gaussian(sigma=0.3))
            do_shoot(spergel2,myImg,"Spergel")

        # Test integrated flux routines against Mathematica
        spergel = galsim.Spergel(nu=nu, scale_radius=1.0)
        np.testing.assert_almost_equal(
            spergel.calculateFluxRadius(1.e-5)/enclosing_radius, 1.0, 4,
            err_msg="Calculated incorrect Spergel(nu={0}) flux-enclosing-radius.".format(nu))
        np.testing.assert_almost_equal(
            spergel.calculateIntegratedFlux(1.e-5)/enclosed_flux, 1.0, 4,
            err_msg="Calculated incorrect Spergel(nu={0}) enclosed flux.".format(nu))

    # Use non-unity values.
    spergel = galsim.Spergel(nu=0.37, flux=1.7, half_light_radius=2.3)

    check_basic(spergel, "Spergel")

    # Check picklability
    check_pickle(spergel, lambda x: x.drawImage(method='no_pixel'))
    check_pickle(spergel)
    check_pickle(galsim.Spergel(0,1))

    # Should raise an exception if both scale_radius and half_light_radius are provided.
    assert_raises(TypeError, galsim.Spergel, nu=0, scale_radius=3, half_light_radius=1)
    assert_raises(TypeError, galsim.Spergel, nu=0)
    assert_raises(TypeError, galsim.Spergel, scale_radius=3)

    # Allowed range = [-0.85, 4.0]
    assert_raises(ValueError, galsim.Spergel, nu=-0.9)
    assert_raises(ValueError, galsim.Spergel, nu=4.1)


@timer
def test_spergel_properties():
    """Test some basic properties of the Spergel profile.
    """
    # Unlike some other similar tests that go back to Gary's original code, this one is based
    # on v2.3.5, the earliest version that was easy for me to go back to and install.
    # I also slightly modified the code for stepk and maxk to use the old _sbp version,
    # since the Python version was already buggy at this point. (Inspired by issue #1324.)

    test_flux = 17.9
    test_hlr = 2.3
    test_scale_radius = 1.6003703593406464
    test_nu = 0.2  # Note: nu<=0 has xValue(0,0) = inf.
    prof = galsim.Spergel(nu=test_nu, half_light_radius=test_hlr, flux=test_flux)

    # Check various properties
    np.testing.assert_equal(prof.centroid, galsim.PositionD(0,0))
    np.testing.assert_almost_equal(prof.maxk, 11.094091231317272)
    np.testing.assert_almost_equal(prof.stepk, 0.2731819698773733)
    np.testing.assert_almost_equal(prof.kValue(0,0), test_flux+0j)
    np.testing.assert_almost_equal(prof.half_light_radius, test_hlr)
    np.testing.assert_almost_equal(prof.xValue(0,0), 2.7808154838921206)
    np.testing.assert_almost_equal(prof.kValue(0,0), (1+0j) * test_flux)
    np.testing.assert_almost_equal(prof.flux, test_flux)
    np.testing.assert_almost_equal(prof.xValue(0,0), prof.max_sb)
    np.testing.assert_almost_equal(prof.scale_radius, test_scale_radius)
    np.testing.assert_almost_equal(prof.half_light_radius, test_hlr)

    # Now create the same profile using scale_radius
    prof = galsim.Spergel(nu=test_nu, scale_radius=test_scale_radius, flux=test_flux)
    np.testing.assert_almost_equal(prof.maxk, 11.094091231317272)
    np.testing.assert_almost_equal(prof.stepk, 0.2731819698773733)
    np.testing.assert_almost_equal(prof.kValue(0,0), test_flux+0j)
    np.testing.assert_almost_equal(prof.half_light_radius, test_hlr)
    np.testing.assert_almost_equal(prof.xValue(0,0), 2.7808154838921206)
    np.testing.assert_almost_equal(prof.kValue(0,0), (1+0j) * test_flux)
    np.testing.assert_almost_equal(prof.flux, test_flux)
    np.testing.assert_almost_equal(prof.xValue(0,0), prof.max_sb)
    np.testing.assert_almost_equal(prof.scale_radius, test_scale_radius)
    np.testing.assert_almost_equal(prof.half_light_radius, test_hlr)

    # Check that stepk and maxk scale correctly with radius
    prof2 = galsim.Spergel(nu=test_nu, half_light_radius=5*test_hlr, flux=test_flux)
    np.testing.assert_almost_equal(prof2.maxk, prof.maxk/5)
    np.testing.assert_almost_equal(prof2.stepk, prof.stepk/5)

    # Check input flux vs output flux
    for inFlux in np.logspace(-2, 2, 10):
        prof3 = galsim.Spergel(nu=test_nu, half_light_radius=test_hlr, flux=inFlux)
        outFlux = prof3.flux
        np.testing.assert_almost_equal(outFlux, inFlux)

    # Repeat with nu < 0
    test_scale_radius2 = 2.1783499913527824
    test_nu2 = -0.2
    prof = galsim.Spergel(nu=test_nu2, half_light_radius=test_hlr, flux=test_flux)

    # Check various properties
    np.testing.assert_equal(prof.centroid, galsim.PositionD(0,0))
    np.testing.assert_almost_equal(prof.maxk, 34.42181161305098)
    np.testing.assert_almost_equal(prof.stepk, 0.23712261251942787)
    np.testing.assert_almost_equal(prof.kValue(0,0), test_flux+0j)
    np.testing.assert_almost_equal(prof.half_light_radius, test_hlr)
    np.testing.assert_almost_equal(prof.xValue(0,0), np.inf)
    np.testing.assert_almost_equal(prof.kValue(0,0), (1+0j) * test_flux)
    np.testing.assert_almost_equal(prof.flux, test_flux)
    np.testing.assert_almost_equal(prof.max_sb, np.inf)
    np.testing.assert_almost_equal(prof.scale_radius, test_scale_radius2)
    np.testing.assert_almost_equal(prof.half_light_radius, test_hlr)

    # Now create the same profile using scale_radius
    prof = galsim.Spergel(nu=test_nu2, scale_radius=test_scale_radius2, flux=test_flux)
    np.testing.assert_almost_equal(prof.maxk, 34.42181161305098)
    np.testing.assert_almost_equal(prof.stepk, 0.23712261251942787)
    np.testing.assert_almost_equal(prof.kValue(0,0), test_flux+0j)
    np.testing.assert_almost_equal(prof.half_light_radius, test_hlr)
    np.testing.assert_almost_equal(prof.xValue(0,0), np.inf)
    np.testing.assert_almost_equal(prof.kValue(0,0), (1+0j) * test_flux)
    np.testing.assert_almost_equal(prof.flux, test_flux)
    np.testing.assert_almost_equal(prof.max_sb, np.inf)
    np.testing.assert_almost_equal(prof.scale_radius, test_scale_radius2)
    np.testing.assert_almost_equal(prof.half_light_radius, test_hlr)

    # Check that stepk and maxk scale correctly with radius
    prof2 = galsim.Spergel(nu=test_nu2, half_light_radius=5*test_hlr, flux=test_flux)
    np.testing.assert_almost_equal(prof2.maxk, prof.maxk/5)
    np.testing.assert_almost_equal(prof2.stepk, prof.stepk/5)


@timer
def test_spergel_radii():
    """Test initialization of Spergel with different types of radius specification.
    """
    import math
    test_spergel_nu = [-0.85, -0.5, 0.0, 0.85, 4.0]
    test_spergel_scale = [20.0, 1.0, 1.0, 0.5, 0.5]
    test_hlr = 1.8

    for nu, scale in zip(test_spergel_nu, test_spergel_scale) :

        test_gal = galsim.Spergel(nu=nu, half_light_radius=test_hlr, flux=1.)

        # Check that the returned half-light radius is correct
        print('test_hlr = ',test_hlr)
        print('test_gal hlr, sr = ',test_gal.half_light_radius,test_gal.scale_radius)
        np.testing.assert_almost_equal(
            test_gal.half_light_radius, test_hlr, decimal=5,
            err_msg = "Error in returned HLR for Spergel HLR constructor, nu=%.1f"%nu)

        # Check that the returned flux is correct
        print('test_gal.flux = ',test_gal.flux)
        np.testing.assert_almost_equal(
            test_gal.flux, 1., decimal=5,
            err_msg = "Error in returned Flux for Spergel HLR constructor, nu=%.1f"%nu)

        # (test half-light radii)
        print('flux = ',test_gal.flux)
        print('hlr = ',test_gal.half_light_radius)
        print('scale = ',test_gal.scale_radius)
        got_hlr = test_gal.half_light_radius
        got_flux = test_gal.flux
        # nu = -0.85 is too difficult to numerically integrate
        if nu > -0.85:
            hlr_sum = radial_integrate(test_gal, 0., got_hlr)
            print('hlr_sum = ',hlr_sum)
            np.testing.assert_almost_equal(
                    hlr_sum, 0.5*got_flux, decimal=4,
                    err_msg = "Error in Spergel half-light radius constructor, nu=%.1f"%nu)

        # Test constructor using scale radius (test scale radius)
        test_gal = galsim.Spergel(nu=nu, scale_radius=scale, flux=1.)

        # Check that the returned scale radius is correct
        print('test_scale = ',scale)
        print('test_gal hlr, sr = ',test_gal.half_light_radius,test_gal.scale_radius)
        np.testing.assert_almost_equal(
            test_gal.scale_radius, scale, decimal=5,
            err_msg = "Error in returned SR for Sersic SR constructor, nu=%.1f"%nu)

        # Check that the returned flux is correct
        print('test_gal.flux = ',test_gal.flux)
        np.testing.assert_almost_equal(
            test_gal.flux, 1., decimal=5,
            err_msg = "Error in returned Flux for Spergel HLR constructor, nu=%.1f"%nu)

        # (test half-light radius)
        got_hlr = test_gal.half_light_radius
        got_flux = test_gal.flux
        # nu = -0.85 is too difficult to numerically integrate
        if nu > -0.85:
            hlr_sum = radial_integrate(test_gal, 0., got_hlr)
            print('hlr_sum = ',hlr_sum)
            np.testing.assert_almost_equal(
                    hlr_sum, 0.5*got_flux, decimal=4,
                    err_msg="Error in HLR for scale_radius constructed Spergel")

        # Check that the getters don't work after modifying the original.
        test_gal_shear = test_gal.shear(g1=0.3, g2=0.1)
        assert_raises(AttributeError, getattr, test_gal_shear, "nu")
        assert_raises(AttributeError, getattr, test_gal_shear, "half_light_radius")
        assert_raises(AttributeError, getattr, test_gal_shear, "scale_radius")


@timer
def test_spergel_flux_scaling():
    """Test flux scaling for Spergel.
    """
    # decimal point to go to for parameter value comparisons
    param_decimal = 12
    test_flux = 17.9
    test_spergel_nu = [-0.85, -0.5, 0.0, 0.85, 4.0]
    test_hlr = 1.8

    # loop through spergel nu
    for test_nu in test_spergel_nu:
        # init with hlr and flux only (should be ok given last tests)
        init_obj = galsim.Spergel(test_nu, half_light_radius=test_hlr, flux=test_flux)

        obj2 = init_obj * 2.
        np.testing.assert_almost_equal(
            init_obj.flux, test_flux, decimal=param_decimal,
            err_msg="Flux param inconsistent after __rmul__ (original).")
        np.testing.assert_almost_equal(
            obj2.flux, test_flux * 2., decimal=param_decimal,
            err_msg="Flux param inconsistent after __rmul__ (result).")

        obj2 = 2. * init_obj
        np.testing.assert_almost_equal(
            init_obj.flux, test_flux, decimal=param_decimal,
            err_msg="Flux param inconsistent after __mul__ (original).")
        np.testing.assert_almost_equal(
            obj2.flux, test_flux * 2., decimal=param_decimal,
            err_msg="Flux param inconsistent after __mul__ (result).")

        obj2 = init_obj / 2.
        np.testing.assert_almost_equal(
             init_obj.flux, test_flux, decimal=param_decimal,
             err_msg="Flux param inconsistent after __div__ (original).")
        np.testing.assert_almost_equal(
            obj2.flux, test_flux / 2., decimal=param_decimal,
            err_msg="Flux param inconsistent after __div__ (result).")


@timer
def test_spergel_05():
    """Test the equivalence of Spergel with nu=0.5 and Exponential
    """
    # cf test_exponential()
    re = 1.0
    r0 = re/1.67839
    # The real value of re/r0 = 1.6783469900166605
    hlr_r0 =  1.6783469900166605
    savedImg = galsim.fits.read(os.path.join(imgdir, "exp_1.fits"))
    dx = 0.2
    myImg = galsim.ImageF(savedImg.bounds, scale=dx)
    spergel = galsim.Spergel(nu=0.5, flux=1., half_light_radius=r0 * hlr_r0)
    spergel.drawImage(myImg, method="sb", use_true_center=False)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using Spergel nu=0.5 disagrees with expected result for Exponential")

    check_basic(spergel, "nu=0.5 Spergel")
    do_kvalue(spergel,myImg,"nu=0.5 Spergel")

    # cf test_exponential_properties()
    test_flux = 17.9
    test_scale = 1.8
    spergel = galsim.Spergel(nu=0.5, flux=test_flux, half_light_radius=test_scale * hlr_r0)
    cen = galsim.PositionD(0, 0)
    np.testing.assert_equal(spergel.centroid, cen)
    np.testing.assert_almost_equal(spergel.kValue(cen), (1+0j) * test_flux)
    np.testing.assert_almost_equal(spergel.flux, test_flux)
    import math
    np.testing.assert_almost_equal(spergel.xValue(cen), 1./(2.*math.pi)*test_flux/test_scale**2,
                                   decimal=5)
    np.testing.assert_almost_equal(spergel.xValue(cen), spergel.max_sb)

    # Also test some random values other than the center:
    expon = galsim.Exponential(flux=test_flux, scale_radius=test_scale)
    for (x,y) in [ (0.1,0.2), (-0.5, 0.4), (0, 0.9), (1.2, 0.1), (2,2) ]:
        pos = galsim.PositionD(x,y)
        np.testing.assert_almost_equal(spergel.xValue(pos), expon.xValue(pos), decimal=5)
        np.testing.assert_almost_equal(spergel.kValue(pos), expon.kValue(pos), decimal=5)


@timer
def test_spergel_shoot():
    """Test Spergel with photon shooting.  Particularly the flux of the final image.
    """
    rng = galsim.BaseDeviate(1234)
    obj = galsim.Spergel(nu=0, half_light_radius=3.5, flux=1.e4)
    im = galsim.Image(100,100, scale=1)
    im.setCenter(0,0)
    added_flux, photons = obj.drawPhot(im, poisson_flux=False, rng=rng.duplicate())
    print('obj.flux = ',obj.flux)
    print('added_flux = ',added_flux)
    print('photon fluxes = ',photons.flux.min(),'..',photons.flux.max())
    print('image flux = ',im.array.sum())
    assert np.isclose(added_flux, obj.flux)
    assert np.isclose(im.array.sum(), obj.flux)
    photons2 = obj.makePhot(poisson_flux=False, rng=rng)
    assert photons2 == photons, "Spergel makePhot not equivalent to drawPhot"

    obj = galsim.Spergel(nu=3.2, half_light_radius=3.5, flux=1.e4)
    added_flux, photons = obj.drawPhot(im, poisson_flux=False, rng=rng.duplicate())
    print('obj.flux = ',obj.flux)
    print('added_flux = ',added_flux)
    print('photon fluxes = ',photons.flux.min(),'..',photons.flux.max())
    print('image flux = ',im.array.sum())
    assert np.isclose(added_flux, obj.flux)
    assert np.isclose(im.array.sum(), obj.flux)
    photons2 = obj.makePhot(poisson_flux=False, rng=rng)
    assert photons2 == photons, "Spergel makePhot not equivalent to drawPhot"

    obj = galsim.Spergel(nu=-0.6, half_light_radius=3.5, flux=1.e4)
    added_flux, photons = obj.drawPhot(im, poisson_flux=False, rng=rng.duplicate())
    print('obj.flux = ',obj.flux)
    print('added_flux = ',added_flux)
    print('photon fluxes = ',photons.flux.min(),'..',photons.flux.max())
    print('image flux = ',im.array.sum())
    assert np.isclose(added_flux, obj.flux)
    assert np.isclose(im.array.sum(), obj.flux)
    photons2 = obj.makePhot(poisson_flux=False, rng=rng)
    assert photons2 == photons, "Spergel makePhot not equivalent to drawPhot"


@timer
def test_ne():
    """Test base.py GSObjects for not-equals."""
    # Define some universal gsps
    gsp = galsim.GSParams(maxk_threshold=1.1e-3, folding_threshold=5.1e-3)

    # Spergel.  Params include nu, half_light_radius, scale_radius, flux, and gsparams.
    # The following should all test unequal:
    gals = [galsim.Spergel(nu=0.0, half_light_radius=1.0),
            galsim.Spergel(nu=0.1, half_light_radius=1.0),
            galsim.Spergel(nu=0.0, half_light_radius=1.1),
            galsim.Spergel(nu=0.0, scale_radius=1.0),
            galsim.Spergel(nu=0.0, half_light_radius=1.0, flux=1.1),
            galsim.Spergel(nu=0.0, half_light_radius=1.0, gsparams=gsp)]
    check_all_diff(gals)


if __name__ == "__main__":
    runtests(__file__)
